
import java.io.*;

class Common
{

  /* rule mode */

  public static final int NEW_MODE = 1;
  public static final int MODIFY_MODE = 2;
  public static final int RETURN_MODE = 3;
  public static final int SELECT_MODE = 4;
  public static final int CHOOSE_MODE = 5;
  public static final int BINARY_MODE = 6;
  public static final int UNARY_MODE = 7;
  public static final int PRIMITIVE_MODE = 8;

  /* nonterminal kind */

  public static final int LABELED = 1;
  public static final int ADD = 2;
  public static final int CALL = 3;
  public static final int TMP = 4;
  public static final int ANONYMOUS = 5;

  /* style options */

  public static final int SPACE = 1;
  public static final int NOSPACE = 2;
  public static final int NEWLINE = 3;
  public static final int EMPTY = 4;
  public static final int INDENT = 5;
  public static final int UNINDENT = 6;

  /* streams */

  public static PrintStream headerStream;
  public static PrintStream cppStream;
  public static PrintStream grammarStream;

  public static PrintStream open (String fileName)
  {
     final int BUF_SIZE = 4096;
     PrintStream stream = null;

     try
     {
        stream =  new PrintStream (
                    new BufferedOutputStream (
                      new FileOutputStream (fileName),
                      BUF_SIZE ));
     }
     catch (FileNotFoundException e)
     {


         err ("File not found: " + fileName);
     }

     return stream;
  }

  public static void close (PrintStream stream)
  {
     if (stream != null)
        stream.close ();
  }

  public static void err (String s)
  {
     System.out.println (s);
     System.err.println (s);
     throw new RuntimeException (s);
  }

  public static void printInfo (String s)
  {
     System.out.println (s);
  }

  /* print */

  private static PrintStream printStream = System.out;
  private static boolean start_line = true;
  public  static int indent = 0;

  public static void switchTo (PrintStream stream)
  {
     printStream = stream;
     indent = 0;
     start_line = true;
  }

  public static void copy (String fileName, String subst)
  {
     final char mark = '%';
     boolean replace = ! empty (subst);

     try
     {
        DataInputStream input =
           new DataInputStream (
              new BufferedInputStream (
                new FileInputStream (fileName) ));

        println ("/* begin of " + fileName + " */");
        println ("");

        while (input.available () > 0)
        {
           String s = input.readLine ();

           if (replace)
           {
              int inx = s.indexOf (mark);
              while (inx >= 0)
              {
                  s = s.substring (0, inx) + subst + s.substring (inx+1);
                  inx = s.indexOf (mark, inx + 1 + subst.length ());
              }
           }

           println (s);
        }

        println ("");
        println ("/* end of " + fileName + " */");
        println ("");

        input.close ();
     }
     catch (FileNotFoundException e)
     {
         err ("File not found: " + fileName);
     }
     catch (IOException e)
     {
         err ("Error reading file: " + fileName);
     }

  }

  private static void print_indentation ()
  {
    if (start_line)
    {
       for (int i = 1; i <= indent; i++)
          printStream.print ("  ");
       start_line = false;
    }
  }

  public static void print (String s)
  {
    print_indentation ();
    printStream.print (s);
  }

  public static void println (String s)
  {
    print_indentation ();
    printStream.println (s);
    start_line = true;
  }

  public static void note (String s)
  {
    println ("/* %% " + s + " */");
    println ("");
  }

  /* strings */

  public static boolean empty (String s)
  {
    return s == null || s.length() == 0;
  }

  public static String strip (String s)
  {
    return s.substring (1, s.length()-1);
  }

  public static boolean isIdentifier (String s)
  {
     boolean b = true;
     int len = s.length ();

     for (int i = 0; b && i < len; i++)
     {
        char c = s.charAt (i);
        if (!( (c >= 'a' && c <= 'z') ||
                (c >= 'A' && c <= 'Z') ||
                (c >= '0' && c <= '(') || (c == '_') )) b = false;
     }

     return b;
  }

}
