
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;

class Decl extends Tool
{

  private FieldInfo defineField (StructInfo target,
                                 String name,
                                 TypeInfo type)
  {
     FieldInfo f = target.lookupField (name);

     if (f == null)
     {
        f = target.createField (name);
        f.type = type;
     }
     else
     {
        if (f.type != type)
           error ("Field " + name + " already defined with different type " +
                  "(" + f.type.ident + " <> " + type.ident + ")");
     }

     return f;
  }

  private void defineEnumItem (EnumInfo target,
                               String name)
  {
     if (! target.containsItem (name))
        target.createItem (name);
  }

  private void defineValue (StructInfo target,
                               String name,
                               String value)
  {
     if (value.equals ("true") || value.equals ("false"))
        defineField (target, name, info.bool_type);
     else
     {
        FieldInfo f = target.lookupField (name);
        if (f == null)
        {
           EnumInfo t = info.lookupEnum (target.ident + "_" + name); // create new type
           f = defineField (target, name, t);
        }
        defineEnumItem (f.type.as_enum(), value);
     }
  }

  private void defineSequence (StructInfo target,
                               StructInfo type)
  {
     StructInfo t = target;
     while (t != null && t.sequence == null)
        t = t.from;

     if (t != null)
     {
        if (t.sequence != type)
           error ("Sequence already defined: " + target.ident);
     }
     else
     {
        target.sequence = type;
     }

     StructInfo u = type.from;
     while (u != null && ! u.sequence_element)
        u = u.from;
     if (u != null)
        error ("Supertype already is sequence element: " +
               type.ident + " (supertype " + u.ident + ")");

     type.sequence_element = true;
  }

  private void putTag (StructInfo type,
                       String tag_name,
                       String tag_value)
  {
     type = type.from;
     if (type == null) error ("Missing super class");

     // search for tag_name in super classes
     FieldInfo f = type.lookupField (tag_name);

     if (f == null)
        // not found => new variable
        defineValue (type, tag_name, tag_value);
     else
        // found => add value
        defineEnumItem (f.type.as_enum(), tag_value);
  }

  private void defineTag (StructInfo target,
                          String tag_name,
                          String tag_value)
  {
     if (! empty (tag_name))
     {

       if (target.tag_name != null && ! target.tag_name.equals (tag_name))
          error ("target " + target.ident + " already have tag " + target.tag_name);

       if (target.tag_name == null)
       {
          target.tag_name = tag_name;
          putTag (target, tag_name, tag_value);
       }

       if (target.tag_value != null && ! target.tag_value.equals (tag_value))
          error ("Type " + target.ident + " already have tag value " + target.tag_value);

       if (target.tag_value == null)
       {
          target.tag_value = tag_value;
       }
     }
  }

  private void checkCompatibility (TypeInfo target, TypeInfo source)
  {
     if (target != source &&
         target instanceof StructInfo &&
         source instanceof StructInfo )
     {
        StructInfo t = (StructInfo) source;
        while (t != null && t != target)
           t = t.from;

        if (t == null)
           error ("Incompatible types: " + target.ident + " and " + source.ident);
     }
  }

  /* enum */

  private EnumInfo act_enum;

  public void openEnum (String name)
  {
     super.openEnum (name);
     act_enum = info.createEnum (name);
  }

  public void closeEnum ()
  {
     super.closeEnum ();
     act_enum = null;
  }

  public void enumItem (String name)
  {
     super.enumItem (name);
     act_enum.createItem (name);
  }

  /* structures */

  private StructInfo act_struct;

  public void openStruct(String type_name)
  {
     super.openStruct (type_name);
     act_struct = info.findStruct (type_name);
  }

  public void closeStruct ()
  {
     super.closeStruct ();
     act_struct = null;
  }

  public void addTag (String type_name,
                      String tag_name,
                      String tag_value)
  {
     super.addTag (type_name, tag_name, tag_value);

     StructInfo t = info.findStruct (type_name);
     defineTag (t, tag_name, tag_value);
  }

  public void addCustom (String type_name, String key)
  {
     super.addCustom (type_name, key);

     if (key.equals ("send"))
     {
        rule.custom_send = true;
     }
     else if (key.equals ("produce"))
     {
        StructInfo t = info.findStruct (type_name);
        t.custom_produce = true;
     }
     else error ("Unknown custom key: " + key);
  }

  public void structField (String name, String type)
  {
     super.structField (name, type);

     FieldInfo f = act_struct.createField (name);
     f.type = info.findType (type);
  }

  /* literals */

  public void literal (String name, String value)
  {
     super.literal (name, value);
     info.storeLiteral (name, value);
  }

  private void check_terminal (String value)
  {
     if (isIdentifier (value))
        info.checkToken (value);
     else
        info.checkLiteral (value);
  }

  public void terminal (String value)
  {
     super.terminal (value);
     check_terminal (value);
  }

  public void predicateTerminal (String value)
  {
     super.predicateTerminal (value);
     check_terminal (value);
  }

  /* rule */

  public void openRule (String name)
  {
    super.openRule (name);
    context.variable_name = "result";
  }

  /* binary */

  public void openBinary (String branch)
  {
     super.openBinary (branch);

     rule.addVariable ("tmp_branch", info.binary_expr_type + " *");
     rule.addVariable ("tmp_right", info.expr_type + " *");
  }

  public void binaryItem (String key, String tag_value)
  {
     super.binaryItem (key, tag_value);

     check_terminal (key);

     StructInfo expr_type = info.findStruct (info.expr_type);
     StructInfo bin_type  = info.findStruct (info.binary_expr_type);
     defineField (bin_type, info.binary_expr_left, expr_type);
     defineField (bin_type, info.binary_expr_right, expr_type);
     defineValue (expr_type, info.expr_tag, tag_value);
  }

  /* unary */

  public void unary (String key, String tag_value, String branch)
  {
     super.unary (key, tag_value, branch);

     rule.unary_tag = tag_value;
     rule.addVariable ("tmp_param", info.expr_type + " *");

     check_terminal (key);

     StructInfo expr_type = info.findStruct (info.expr_type);
     StructInfo unary_type = info.findStruct (info.unary_expr_type);
     defineField (unary_type, info.unary_expr_param, expr_type);
     defineValue (expr_type, info.expr_tag, tag_value);
  }

  /* blocks */

  public void openBlock ()
  {
     super.openBlock ();
  }

  public void closeBlock ()
  {
     super.closeBlock ();
  }

  /* alternatives */

  public void openAlternative ()
  {
     super.openAlternative ();
  }

  public void closeAlternative ()
  {
     super.closeAlternative ();
  }

  /* nonterminals */

  public void nonterminal (int kind, String label, String name)
  {
     super.nonterminal (kind, label, name);

     RuleNode r = info.findRule (name); // sub-rule
     int m = r.mode;  // sub-rule mode
     TypeInfo t = r.type; // sub-rule type

     // if (t instanceof StructInfo)
     //    ((StructInfo) t).used_as_field = true;
     t.used_as_field = true;

     // check sub-rule mode
     if (kind == CALL)
     {
        if (m != MODIFY_MODE)
           error ("Subrule with <modify> parameter required");
     }
     else
     {
        if (m == MODIFY_MODE)
           error ("Subrule with <modify> parameter is not allowed here");
     }

     if (kind == LABELED)
     {
        current.field = defineField (context.type, label, t);
        current.variable = rule.addVariable ("tmp_" + label, t.decl());

        if (t instanceof StructInfo)
           alt.addRefKey (label);
        else
           alt.addStrKey (label);
     }
     else if (kind == ADD)
     {
        current.variable = rule.addVariable ("item", t.decl());
        defineSequence (context.type, t.as_struct());
        context.loop = true;
        block.loop = true;
        alt.loop = true;
     }
     else if (kind == CALL)
     {
        checkCompatibility (t, context.type); // assignment to sub-rule parameter
     }
     else if (kind == TMP)
     {
        rule.temp_type = t; // store <tmp> type
     }
     else if (kind == ANONYMOUS)
     {
        checkCompatibility (rule.type, t); // assignment to rule result
        if (rule.mode == SELECT_MODE)
        {
           alt.selection_rule = r;
           alt.selection_type = t.as_struct(); // type of this selection alternative
        }
     }
     else error ("Internal error");
  }

  /* options */

  public void resultOption (String type_name)
  {
     super.resultOption (type_name);

     StructInfo t = info.findStruct (type_name);

     context.variable_name = rule.addVariable ("result"+context.suffix(), t.decl());

     checkCompatibility (rule.type, t); // assignment to rule result
     context.type = t; // set act_type
  }

  public void storeOption (String name)
  {
     super.storeOption (name);

     TypeInfo t = (rule.temp_type != null) ? rule.temp_type : rule.type;
     defineField (context.type, name, t);
  }

  public void setOption (String name, String value)
  {
     super.setOption (name, value);
     defineValue (context.type, name, value);
     alt.addOrdKey (name, value);
  }

  /* output */

  public void sendLiterals ()
  {
     Enumeration seq = info.getLiterals ();
     while (seq.hasMoreElements ())
     {
        String s;
        DictItem2 item = (DictItem2) seq.nextElement ();

        if (item.value.length () == 1)
           s = '\'' + item.value + '\'';
        else
           s = '\"' + item.value + '\"';

        println (item.ident);
        println ("  options { paraphrase=\"\'" + item.value + "\'\"; }");
        println ("  : " + s + " ;");
        println ("");
     }
  }

  public void sendTokens ()
  {
     Enumeration seq = info.getTokens ();

     println ("tokens");
     println ("{");
     indent ++;

     while (seq.hasMoreElements ())
     {
        DictItem item = (DictItem) seq.nextElement ();
        println ('\"' + item.ident + '\"' + ";");
     }

     indent --;
     println ("}");
     println ("");
  }

  private void send_enumerations ()
  {
     Enumeration seq = info.getTypes ();
     while (seq.hasMoreElements ())
     {
        TypeInfo t = (TypeInfo) seq.nextElement ();
        if (t instanceof EnumInfo)
        {
           EnumInfo e = (EnumInfo) t;

           println ("enum " + e.ident);
           println ("{");
           indent ++;

           Enumeration items = e.getItems ();
           while (items.hasMoreElements ())
           {
              DictItem item = (DictItem) items.nextElement ();
              print (item.ident);

              if (items.hasMoreElements ()) print (", ");
              println ("");
           }

           indent --;
           println ("};");
           println ("");
        }
     }
  }

  private void send_basic_type ()
  {
     println ("struct " + info.basic_type);
     println ("{");
     indent ++;
     println ("virtual string produce ();");
     println ("virtual void produce_attributes () {}");
     println ("virtual void produce_sequence () {}");
     println ("virtual string getTypeName () { return " + '"' + info.basic_type + '"' + "; }");
     println ("virtual bool isSimple () { return true; }");
     println ("virtual bool haveSubitems () { return false; }");
     indent --;
     println ("};");
     println ("");
  }

  private void send_structure (StructInfo s)
  {
     if (! s.declared)
     {
        if (s.from != null)
           send_structure (s.from);
        s.declared = true;

        print ("struct " + s.ident + " : public " );
        if (s.from != null)
           print (s.from.ident);
        else
           print (info.basic_type);
        println ("");

        println ("{");
        indent ++;

        // sequences
        s.any_sequence = (s.sequence != null);
        if (s.from != null && s.from.any_sequence)
           s.any_sequence = true;

        // parameters
        if (s.from != null)
          s.copyParams (s.from);

        boolean any = false;

        // fields
        Enumeration fields = s.getFields ();
        while (fields.hasMoreElements ())
        {
           FieldInfo f = (FieldInfo) fields.nextElement ();
           println (f.type.decl() + " " + f.ident + ";");
           // if (s.tag_name == null || f.ident.compareTo (s.tag_name) != 0)
              s.addParam (f);
           any = true;
        }

        if (any) println ("");
        any = false;

        if (s.sequence_element)
        {
           println (s.ident + " * next;");
           any = true;
        }

        if (s.sequence != null)
        {
           println (s.sequence.ident + " * first;");
           println (s.sequence.ident + " * last;");

           println ("void add (" + s.sequence.ident + " * item);");
           any = true;
        }

        if (any) println ("");

        // constructor
        println (s.ident + " ();");

        // constructor with parameters
        int cnt = s.getParamCount ();
        s.create_func = (cnt <= StructInfo.param_limit && cnt != 0);

        if (s.create_func)
        {
           // print ("%%% ");
           if (cnt == 1) print ("explicit ");
           print (s.ident + " (");
           for (int i = 0; i < cnt; i++)
           {
              if (i > 0) print (", ");
              FieldInfo f = s.getParam (i);
              print (f.type.decl() + " p_" + f.ident);
           }
           println (");");
        }

        // produce
        println ("");

        if (s.custom_produce)
           println ("virtual string produce ();");
           /* produce_attributes and produce_sequence
              are generated even if s.custom_produce == true */

        if (s.getFieldCount () != 0)
           println ("virtual void produce_attributes ();");

        if (s.sequence != null)
            println ("virtual void produce_sequence ();");

        println ("virtual string getTypeName () { return " + '"' + s.ident + '"' + "; }");

        if (s.custom_produce ||
            (cnt > StructInfo.param_limit &&
             (s.from == null ||
              s.from.getParamCount () <= StructInfo.param_limit)))
           println ("virtual bool isSimple () { return false; }");

        if (s.sequence != null)
           println ("virtual bool haveSubitems () { return first != NULL; }");

        indent --;
        println ("};");
        println ("");
     }
  }

  private void send_init (String name, String value)
  {
     println (",");
     print (name + " (" + value + ")");
  }

  private void send_cont (StructInfo s)
  {
     if (s.sequence != null)
     {
        send_init ("first", "NULL");
        send_init ("last", "NULL");
     }

     if (s.sequence_element)
        send_init ("next", "NULL");

     println ("");
     indent --;
     print ("{ ");
     if (s.tag_name != null)
     {
        print (s.tag_name + " = " + s.tag_value + "; ");
     }
     println ("}");
     println ("");
  }

  private void send_methods (StructInfo s)
  {
     // add method
     if (s.sequence != null)
     {
        println ("void " + s.ident + "::add (" + s.sequence.ident + " * item)");
        println ("{");
        indent ++;

        println ("if (first == NULL)");
        indent ++;
        println ("first = item;");
        indent --;
        println ("else");
        indent ++;
        println ("last->next = item;");
        indent --;

        println ("last = item;");
        println ("item->next = NULL;");

        indent --;
        println ("}");
        println ("");
     }

     // constructor
     println (s.ident + "::" + s.ident + " () :");
     indent ++;

     if (s.from == null)
        print (info.basic_type + " ()");
     else
        print (s.from.ident + " ()");

     Enumeration fields = s.getFields ();
     while (fields.hasMoreElements ())
     {
        String txt = "";
        FieldInfo f = (FieldInfo) fields.nextElement ();

        if (f.type instanceof StructInfo) txt = "NULL";
        else if (f.type instanceof EnumInfo) txt = f.type.ident + " (0)";
        else if (f.type == info.bool_type) txt = "false";
        else if (f.type == info.int_type) txt = "0";
        else if (f.type == info.string_type) txt = "\"\"";
        else error ("Internal error");

        send_init (f.ident, txt);
     }

     send_cont (s);

     // constructor with parameters
     if (s.create_func)
     {
        print (s.ident + "::" + s.ident + " (");
        int cnt = s.getParamCount ();
        for (int i = 0; i < cnt; i++)
        {
           if (i > 0) print (", ");
           FieldInfo f = s.getParam (i);
           print (f.type.decl() + " p_" + f.ident);
        }
        println (") :");
        indent ++;

        // parent
        int super_cnt = 0;
        if (s.from == null)
           print (info.basic_type + " ()");
        else
        {
           print (s.from.ident);
           print (" (");
           super_cnt = s.from.getParamCount ();
           for (int i = 0; i < super_cnt; i++)
           {
              if (i > 0) print (", ");
              FieldInfo f = s.from.getParam (i);
              print ("p_" + f.ident);
           }
           print (")");
        }

        // own
        for (int i = super_cnt; i < cnt; i++)
        {
           FieldInfo f = s.getParam (i);
           send_init (f.ident, "p_" + f.ident);
        }

        send_cont (s);
     }
  }

  private void send_structures ()
  {
     println ("struct " + info.basic_type + ";");

     Enumeration seq = info.getTypes ();
     while (seq.hasMoreElements ())
     {
        TypeInfo t = (TypeInfo) seq.nextElement ();
        if (t instanceof StructInfo)
        {
           StructInfo s = (StructInfo) t;
           println ("struct " + s.ident + ";");
        }
     }
     println ("");

     send_basic_type ();

     seq = info.getTypes ();
     while (seq.hasMoreElements ())
     {
        TypeInfo t = (TypeInfo) seq.nextElement ();
        if (t instanceof StructInfo)
        {
           StructInfo s = (StructInfo) t;
           send_structure (s);
        }
     }
  }

  private void send_function_headers ()
  {
     Enumeration seq = info.getRules ();
     while (seq.hasMoreElements ())
     {
        RuleInfo r = (RuleInfo) seq.nextElement ();
        RuleNode n = r.node;
        println ("void send_" + n.name + " (" + n.type.decl() + " item" + ");");
     }
  }

  public void sendHeader ()
  {
     send_enumerations ();
     send_structures ();
     send_function_headers ();
  }

  public void sendCode ()
  {
     Enumeration seq = info.getTypes ();
     while (seq.hasMoreElements ())
     {
        TypeInfo t = (TypeInfo) seq.nextElement ();
        if (t instanceof StructInfo)
        {
           StructInfo s = (StructInfo) t;
           send_methods (s);
        }
     }
  }

}

