
import java.util.Enumeration;

class Dupl extends Tool
{

  /* rules */

  public void openRule (String name)
  {
    super.openRule (name);
    println (name);
    indent ++;
  }

  private void ruleVariables ()
  {
    Enumeration seq = rule.getVariables ();
    if (seq.hasMoreElements ())
    {

       println ("{");
       indent ++;

       while (seq.hasMoreElements ())
       {
          VarInfo item = (VarInfo) seq.nextElement ();
          println (item.type + " " + item.ident + ";");
       }

       indent --;
       println ("}");
    }
  }

  public void ruleParam (int mode,
                         String type_name)
  {
    super.ruleParam (mode, type_name);

    if (mode != MODIFY_MODE)
       print ("returns ");

    println ("[ " + type_name + " * result ]");
    ruleVariables ();
    println (":");

    if (mode == NEW_MODE)
    {
       indent ++;
       println ("{ result = new " + type_name + " (); }");
       indent --;
    }

  }

  public void closeRule ()
  {
    super.closeRule ();
    println (";");
    indent --;
    println ("");
  }

  /* binary */

  private String binary_branch;
  private int binary_item_count;

  public void openBinary (String branch)
  {
     super.openBinary (branch);

     println ("returns [ " + info.expr_type + " * result ]");
     ruleVariables ();
     println (":");
     indent ++;
     println ("result=" + branch);

     println ("(");
     binary_branch = branch;
     binary_item_count = 0;
  }

  public void binaryItem (String key, String tag_value)
  {
     super.binaryItem (key, tag_value);

     binary_item_count ++;
     if (binary_item_count != 1) println ("|");
     indent ++;

     put_terminal (key);
     println ("");

     String v = "tmp_branch";
     String t = "tmp_right";

     println ("{ " + v + " = new " + info.binary_expr_type + " ();");
     println ("  " + v + "->" + info.expr_tag + " = " + tag_value +";");
     println ("  " + v + "->" + info.binary_expr_left + " = result; }");

     print   (t + " = " + binary_branch + " ");
     println ("{ " + v + "->" + info.binary_expr_right + " = " + t + "; }");

     println ("{ result = " + v + "; }");

     indent --;
  }

  public void closeBinary ()
  {
     super.closeBinary ();
     println (")*");
     indent --;
  }

  /* unary */

  public void unary (String key, String tag_value, String branch)
  {
     super.unary (key, tag_value, branch);

     println ("returns [ " + info.unary_expr_type + " * result ]");
     ruleVariables ();
     println (":");
     indent ++;
     println ("{ result = new " + info.unary_expr_type + " (); }");

     put_terminal (key);
     println ("");

     println ("{ result->" + info.expr_tag + " = " + tag_value +"; }");

     print   ("tmp_param=" + branch + " ");
     println ("{ result->" + info.unary_expr_param + "=tmp_param; }");

     indent --;
  }

  /* primitive */

  private int primitive_item_count;

  public void openPrimitive ()
  {
     super.openPrimitive ();

     println ("returns [ string result ]");
     println (":");

     primitive_item_count = 0;
  }

  public void primitiveItem (String branch)
  {
     super.primitiveItem (branch);

     primitive_item_count ++;
     if (primitive_item_count != 1) print ("| ");
     indent ++;

     String id = "tmp_" + primitive_item_count;
     print (id + ":" + branch);
     println (" { result=" + id + "->getText(); }");

     indent --;
  }

  /* block */

  public void openBlock ()
  {
    super.openBlock ();
    if (block.above != null) println ("(");
  }

  public void closeBlock ()
  {
    super.closeBlock ();

    if (block.above != null)
    {
       print (")");
       if (block.mark != ' ') print (""+block.mark);
       println ("");
    }
  }

  /* alternative */

  public void openAlternative ()
  {
    super.openAlternative ();

    if (alt.index != 1) println ("|");
    indent ++;
  }

  public void closeAlternative ()
  {
    super.closeAlternative ();

    if (is_choose_block ())
    {
       println ("{ result = " + context.variable_name + "; }");
    }
    indent --;
  }

  /* terminal */

  private void put_terminal (String value)
  {
    if (isIdentifier (value))
       print ('\"' + value + '\"');
    else
       print (info.findLiteral (value));
  }

  public void terminal (String value)
  {
    super.terminal (value);
    put_terminal (value);
    println ("");
  }

  /* nonterminal */

  public void nonterminal (int kind, String label, String name)
  {
    super.nonterminal (kind, label, name);

    TypeInfo t = info.findRule (name).type;

    if (kind == LABELED)
    {
       String v = current.variable;
       print (v + "=" + name);
       println (" { " + context.variable_name + "->" + label + "=" + v +"; }");
    }
    else if (kind == ADD)
    {
       String v = current.variable;
       print (v + "=" + name);
       println (" { " + context.variable_name + "->add (" + v + "); }");
    }
    else if (kind == CALL)
    {
       println (name + " [ " + context.variable_name + " ]");
    }
    else if (kind == TMP)
    {
       print ("{ " + t.decl () + " tmp; } ");
       println ("tmp=" + name);
    }
    else if (kind == ANONYMOUS)
    {
       println (context.variable_name + "=" + name);
    }
    else error ("Internal error");
  }

  /* predicate */

  public void openPredicate ()
  {
     super.openPredicate ();
     print ("( ");
  }

  public void closePredicate ()
  {
     super.closePredicate ();
     print (") => ");
  }

  public void predicateNonterminal (String name)
  {
     super.predicateNonterminal (name);
     print (name + " ");
  }

  public void predicateTerminal (String value)
  {
     super.predicateTerminal (value);
     put_terminal (value);
  }

  /* result option */

  public void resultOption (String type_name)
  {
     super.resultOption (type_name);

     println ("{ " + context.variable_name + " = new " + type_name + " (); }");
  }

  /* store option */

  public void storeOption (String name)
  {
    super.storeOption (name);

    String v =  (rule.temp_type != null) ? "tmp" : "result" ;
    println ("{ " + context.variable_name + "->" + name + " = " + v + "; }");
  }

  /* set option */

  public void setOption (String name, String value)
  {
    super.setOption (name, value);
    println ("{ " + context.variable_name + "->" + name + " = " + value + "; }");
  }

  /* silent option */

  public void silentOption ()
  {
    super.silentOption ();
    println ("options { warnWhenFollowAmbig = false; }:");
  }
}
