
class List extends Tool
{

  private String cast (String type, String expr)
  {
    return "dynamic_cast <" + type + "*> (" + expr + ")";
  }

  private void check (String expr)
  {
    println ("assert (" + expr + ");");
  }

  private void openContext ()
  {
    if (context.loop)
    {
       println (context.type.sequence.ident + " * " + context.subitem_name () +
                " = " + context.item_name() + "->first;");

       if (alt != null)
          alt.end_loop = true;
    }
  }

  private void closeLoop ()
  {
    check (context.subitem_name () + " == NULL");
  }

  /* rules */

  public void openRule (String name)
  {
    super.openRule (name);

    if (rule.custom_send) println ("/*");
    println ("void send_" + name + " (" + rule.type.decl() + " item" + ")");
    println ("{");
    indent ++;

    if (rule.type instanceof StructInfo) check ("item != NULL");
    openContext ();
  }

  public void ruleParam (int mode,
                         String type_name)
  {
    super.ruleParam (mode, type_name);
  }

  public void closeRule ()
  {
    super.closeRule ();

    if (context.loop)
       closeLoop ();

    indent --;
    println ("}");
    if (rule.custom_send) println ("*/");
    println ("");
  }

  /* binary */

  private RuleNode binary_branch;
  private int      binary_count;

  public void openBinary (String branch)
  {
     super.openBinary (branch);
     binary_branch = info.findRule (branch);
     binary_count = 0;
  }

  public void binaryItem (String key, String tag_value)
  {
     super.binaryItem (key, tag_value);

     binary_count ++;
     if (binary_count > 1) print ("else ");

     println ("if (item->" + info.expr_tag + " == " + tag_value + ")");
     println ("{");
     indent ++;

     println (info.binary_expr_type + " * item1 = " +
              cast (info.binary_expr_type, "item") + ";" );

     check ("item1 != NULL");
     put (rule, rule.type, "item1->" + info.binary_expr_left);
     put_terminal (key);
     put (binary_branch, rule.type, "item1->" + info.binary_expr_right);

     indent --;
     println ("}");
  }

  public void closeBinary ()
  {
     super.closeBinary ();

     print ("else ");
     put (binary_branch, rule.type, "item");
  }

  /* unary */

  public void unary (String key, String tag_value, String branch)
  {
     super.unary (key, tag_value, branch);

     check ("item->" + info.expr_tag + " == " + tag_value);
     put_terminal (key);

     RuleNode r = info.findRule (branch);
     put (r, rule.type, "item->" + info.unary_expr_param);
  }

  /* primitive */

  public void openPrimitive ()
  {
     super.openPrimitive ();
     println ("send (item);");
  }

  /* block */

  public void openBlock ()
  {
    super.openBlock ();

    if (! is_select_block () && ! is_choose_block ())
       if (block.mark == '+' || block.mark =='*')
          if (block.loop)
          {
             if (block.count > 1) error ("Too complicated loop");
             println ("while (" + context.subitem_name() + " != NULL)");
             block.braces = true;
             block.subitem_loop = true; /* no condtition for inner alternative */
          }

    if (block.braces)
    {
       println ("{");
       indent ++;
    }
  }

  public void closeBlock ()
  {
    super.closeBlock ();

    if (block.braces)
    {
       indent --;
       println ("}");
    }

    if (is_select_block ())
    {
       println ("else assert (false);");
    }
    else if (is_choose_block ())
    {
       if (block.mark == ' ')
          println ("else assert (false);");
       else
       {
          print ("else ");
          put (rule.stored_rule,
               context.type,
               context.item_name());
       }
    }
    else if (block.mark == ' ' && block.count != 1)
    {
       println ("else assert (false);");
    }
  }

  /* alternative */

  private void condition ()
  {
     int count = 0;

     int len = alt.ordKeys.size ();
     for (int i=0; i<len; i++)
     {
        DictItem2 item = (DictItem2) alt.ordKeys.get (i);
        count ++;
        if (count > 1) print (" && ");
        print (context.item_name() + "->" + item.ident + " == " + item.value);
     }

     len = alt.refKeys.size ();
     for (int i=0; i<len; i++)
     {
        String id = (String) alt.refKeys.get (i);
        count ++;
        if (count > 1) print (" && ");
        print (context.item_name() + "->" + id + " != NULL");
     }

     len = alt.strKeys.size ();
     for (int i=0; i<len; i++)
     {
        String id = (String) alt.strKeys.get (i);
        count ++;
        if (count > 1) print (" && ");
        print ("! " + context.item_name() + "->" + id + ".empty()");
     }

     if (count == 0)
        error ("No condition for this alternative");
        // print ("COND");
  }

  public void openAlternative ()
  {
    super.openAlternative ();

    if (! block.subitem_loop)
       if (block.mark != ' ' || block.count != 1)
       {
          if (alt.index != 1) print ("else ");
          print ("if (");

          if (is_select_block ())
          {
             print (cast (alt.selection_type.ident, "item") + " != NULL");
             if (alt.selection_rule.mode == UNARY_MODE)
                print (" && " + cast (alt.selection_type.ident, "item") + "->" + info.expr_tag +
                       " == " + alt.selection_rule.unary_tag);
          }
          else if (is_choose_block ())
             print (cast (context.type.ident, "item") + " != NULL");
          else if (alt.loop)
             print (context.subitem_name() + " != NULL");
          else
             condition ();

          println (")");

          alt.braces = true;
       }

    if (alt.braces)
    {
       println ("{");
       indent ++;
    }
  }

  public void closeAlternative ()
  {
    super.closeAlternative ();

    if (alt.end_loop)
       closeLoop ();

    if (alt.braces)
    {
       indent --;
       println ("}");
    }
  }

  /* terminals */

  private void put_terminal (String value)
  {
    println ("send (" + '\"' + value + '\"' +  ");");
  }

  public void terminal (String name)
  {
    super.terminal (name);
    put_terminal (name);
  }

  private void put (RuleNode subrule,
                    TypeInfo expr_type,
                    String expr)
  {
     String txt = expr;
     if (subrule.type != expr_type)
        txt = cast (subrule.type.ident, txt);
     println ("send_" + subrule.name + " (" + txt + ");");
  }

  /* nonterminals */

  public void nonterminal (int kind, String label, String name)
  {
    super.nonterminal (kind, label, name);

    RuleNode r = info.findRule (name);

    if (kind == TMP || kind == ANONYMOUS && rule.mode == CHOOSE_MODE)
    {
       rule.stored_rule = r;
    }
    else
    {
       if (kind == LABELED)
       {
          put (r, current.field.type,
                  context.item_name () + "->" + label);
       }
       else if (kind == ADD)
       {
          String id = context.subitem_name ();
          put (r, context.type.sequence, id);
          println (id + " = " + id + "->next;");
       }
       else if (kind == CALL)
       {
          put (r, context.type, context.item_name ());
       }
       else if (kind == ANONYMOUS)
       {
          if (rule.mode == SELECT_MODE)
          {
             put (r, rule.type, context.item_name ());
          }
          else if (rule.mode == RETURN_MODE)
          {
             put (r, rule.type, context.item_name ());
          }
          else error ("Internal error");
       }
       else error ("Internal error");
    }
  }

  /* options */

  public void resultOption (String type_name)
  {
    super.resultOption (type_name);

    println (type_name + " * " + context.item_name () +
             " = " + cast (type_name, "item") + ";" );

    openContext ();
  }

  public void storeOption (String name)
  {
    super.storeOption (name);

    FieldInfo field = context.type.findField (name);

    RuleNode r;

    if (block.mark == '*')
       r = rule; /* again */
    else
       r = rule.stored_rule; /* stored rule */

    put (r,
         field.type,
         context.item_name() + "->" + name);
  }

  public void setOption (String name, String value)
  {
    super.setOption (name, value);
    check (context.item_name () + "->" + name + " == " + value);
  }

  public void styleOption (int style)
  {
     super.styleOption (style);

     switch (style)
     {
        case SPACE:    println ("style_space ();");    break;
        case NOSPACE:  println ("style_nospace ();");   break;
        case NEWLINE:  println ("style_newline ();");  break;
        case EMPTY:    println ("style_empty ();");  break;
        case INDENT:   println ("style_indent ();");   break;
        case UNINDENT: println ("style_unindent ();"); break;
     }
  }

}
