
import java.io.*;
import antlr.collections.AST;

class Main extends Common
{

  public static ParserInfo info;

  /* print tree */

  public static void printTree (AST branch)
  {
    while (branch != null)
    {
       print (branch.getText ());
       print (" (" + branch.getType () + ")");
       print (" (" + branch.getClass().getName() + ")");
       println ("");

       indent ++;

       if (branch instanceof BlockNode)
          println ("mark = " + ((BlockNode) branch).mark);

       printTree (branch.getFirstChild ());
       indent --;

       branch = branch.getNextSibling ();
    }
  }

  /* process */

  private static EGLexer lexer;
  private static EGParser parser;

  private static void process (Tool tool) throws Exception
  {
    printInfo (tool.getClass().getName ());
    tool.info = info;

 	  EGTreeParser tree_parser = new EGTreeParser ();
	  tree_parser.tool = tool;
 		tree_parser.grammar (parser.getAST ());

 		Tool.first_pass = false;



  }

  /* main */

  public static void main (String[] args)
  {
		try
		{
		  try
		  {
     		// lexer and parser
   			lexer = new EGLexer (new DataInputStream (System.in));
   			parser = new EGParser (lexer);
   			parser.setASTFactory (new NodeFactory ());
   			parser.grammar ();

   			if (false)
   			{
   			    // print tree
      		  AST ast = parser.getAST ();
      		  println ("tree:");
      		  printTree (ast);
      		  println ("");
   		  }

   		  // dictionary of rules, structures, enumerations
     		info = new ParserInfo ();

     		// test tree parser
     		// process (new Test ());

        // collect names of rules and structures
     		process (new Scanner ());

     		final String dir = "./";          // target directory and file name prefix
     		final String prod = info.product_name; // output file name prefix
     		final String padding = info.padding;   // prefix of input files
     		final String util_header = info.util_header; // header file with utility functions

     		if (empty (prod)) err ("Missing product_name option");
     		if (empty (padding)) err ("Missing padding option");
     		if (empty (util_header)) err ("Missing padding option");

		    // open streams
     		headerStream  = open (dir + prod + ".hpp");
     		cppStream     = open (dir + prod + ".cpp");
     		grammarStream = open (dir + prod + ".g");

     		// fields, sequences, enumerations, tokens, literals, type compatibility
     		Decl decl = new Decl ();
   		  process (decl);

   		  // struct and enum declarations
   		  switchTo (headerStream);
   		  String sym = "__" + prod + "_HPP__";
        println ("#ifndef " + sym);
        println ("#define " + sym);
        println ("");

   		  decl.sendHeader ();

   		  /*
   		  println ("");
   		  println ("void send (string s);");
   		  println ("void style_space ();");
   		  println ("void style_nospace ();");
   		  println ("void style_newline ();");
   		  println ("void style_empty ();");
   		  println ("void style_indent ();");
   		  println ("void style_unindent ();");
   		  println ("");
   		  println ("void indent ();");
   		  println ("void unindent ();");
   		  println ("");
   		  println ("string find_ident (string s);");
   		  println ("void init (string name, string type);");
   		  println ("void store (string target, string field, string value);");
   		  println ("void insert (string target, string value);");
        println ("");
   		  println ("void open (string name, string type);");
   		  println ("void param (string value);");
   		  println ("void close ();");
        println ("");
   		  println ("string conv_string (string value);");
   		  println ("string conv_bool (bool value);");
        println ("");
        */

        println ("");
        println ("#endif /* " + sym + " */");

   		  // write parser
   		  switchTo (grammarStream);
   		  println ("header");
   		  println ("{");
   		  println ("  #include " + '"' + prod + ".hpp" + '"');
   		  println ("}");
   		  println ("");
   		  copy (padding + "1.g", prod);
   		  decl.sendTokens ();

   		  // duplicate grammar rules
   		  Dupl dupl = new Dupl ();
   		  process (dupl);

   		  // write lexer
   		  copy (padding + "2.g", prod);
   		  decl.sendLiterals ();
   		  copy (padding + "3.g", "");

   		  // C++ code
   		  switchTo (cppStream);
   		  println ("#include " + '<' + "string" + '>');
   		  println ("#include " + '"' + prod + ".hpp" + '"');
   		  println ("#include " + '"' + util_header + '"');
   		  println ("");

   		  note ("decl");
   		  decl.sendCode ();

   		  note ("list");
   		  process (new List ());

   		  Dump dump = new Dump ();
   		  dump.info = info;
   		  note ("dump");
   		  dump.sendCode ();
   		}
   		finally
   		{
   		  close (headerStream);
   		  close (cppStream);
   		  close (grammarStream);
   		}

      printInfo ("O.K.");
			System.err.println ("O.K.");
		}
		catch(Exception e)
		{
			System.err.println("exception: " + e);
			e.printStackTrace (System.err);
		}
	}

}
