
import java.util.Enumeration;

class ParserInfo extends Common
{
  /* configuration */

  String product_name;
  String padding;

  String util_header = "utility.h";

  String basic_type = "_TBasic";

  String expr_type = "_TExpr";
  String expr_tag  = "_tag";

  String binary_expr_type  = "_TBinaryExpr";
  String binary_expr_left  = "_left";
  String binary_expr_right = "_right";

  String unary_expr_type   = "_TUnaryExpr";
  String unary_expr_param  = "_param";

  /* types */

  private DictTable type_list = new DictTable ();

  public TypeInfo getType (String name)
  {
    return (TypeInfo) type_list.get (name);
  }

  public TypeInfo findType (String name)
  {
     TypeInfo t = getType (name);
     if (t == null)
        err ("Undefined type: " + name);
     return t;
  }

  private void createType (TypeInfo type)
  {
    if (getType (type.ident) != null)
       err ("Type already defined: " + type.ident);

    type_list.store (type);
  }

  public Enumeration getTypes ()
  {
    return type_list.elements ();
  }

  /* simple types */

  public final TypeInfo string_type = newType ("string");
  public final TypeInfo bool_type   = newType ("bool");
  public final TypeInfo int_type    = newType ("int");

  private TypeInfo newType (String name)
  {
    TypeInfo t = new TypeInfo ();
    t.ident = name;
    createType (t);
    return t;
  }

  /* enums */

  public EnumInfo getEnum (String name)
  {
    TypeInfo t = getType (name);

    if (t != null && ! (t instanceof EnumInfo))
       err ("Enum type required: " + name);



    return (EnumInfo) t;
  }

  public EnumInfo findEnum (String name)
  {
     EnumInfo t = getEnum (name);
     if (t == null)
        err ("Undefined type: " + name);
     return t;
  }

  public EnumInfo createEnum (String name)
  {
    EnumInfo typ = new EnumInfo ();
    typ.ident = name;
    createType (typ);
    return typ;
  }

  public EnumInfo lookupEnum (String name)
  {
     EnumInfo t = getEnum (name);
     if (t == null)
        t = createEnum (name);
     return t;


  }

  /* structures */

  public StructInfo getStruct (String name)
  {
    TypeInfo t = getType (name);

    if (t != null && ! (t instanceof StructInfo))
       err ("Struct type required: " + name);

    return (StructInfo) t;
  }

  public StructInfo findStruct (String name)
  {
     StructInfo t = getStruct (name);
     if (t == null)
        err ("Undefined type: " + name);
     return t;
  }

  public StructInfo createStruct (String name)
  {
    StructInfo t = new StructInfo ();
    t.ident = name;
    createType (t);
    return t;
  }

  public StructInfo lookupStruct (String name)
  {
     StructInfo t = getStruct (name);
     if (t == null)
        t = createStruct (name);
     return t;
  }

  /* literals */

  private DictTable2 literal_list  = new DictTable2 ();

  public String getLiteral (String value)
  {
    DictItem2 item = literal_list.searchValue (value);
    if (item == null)
       return null;
    else
       return item.ident;
  }

  public String findLiteral (String value)
  {
    String s = getLiteral (value);
    if (empty (s))
       err ("Undefined literal: " + value);
    return s;
  }

  public void storeLiteral (String name, String value)
  {
    if (empty (name))
       err ("Literal without name: " + value);

    if (empty (value))
       err ("Literal without value: " + name);

    if (literal_list.containsKey (name))
       err ("Literal identifier already defined: " + name);

    if (literal_list.containsValue (value))
       err ("Literal value already defined: " + value);


    DictItem2 item = new DictItem2 ();
    item.ident = name;
    item.value = value;
    literal_list.store (item);
  }

  public void checkLiteral (String value)
  {
    if (! literal_list.containsValue (value))
    {
       String name = "LITERAL_" + literal_list.size ();
       storeLiteral (name, value);
    }
  }

  public Enumeration getLiterals ()
  {
    return literal_list.elements ();
  }

  /* tokens */

  private DictTable token_list = new DictTable ();

  public boolean containsToken (String name)
  {
    return token_list.contains (name);
  }

  public void checkToken (String name)
  {
    if (! token_list.contains (name))
    {
       DictItem item = new DictItem ();
       item.ident = name;
       token_list.store (item);
    }
  }

  public Enumeration getTokens ()
  {
    return token_list.elements ();
  }

  /* rules */

  private DictTable rule_list = new DictTable ();

  public RuleNode getRule (String name)
  {
    RuleInfo item = (RuleInfo) rule_list.get (name);
    if (item == null)
       return null;
    else
       return item.node;
  }

  public RuleNode findRule (String name)
  {
     RuleNode r = getRule (name);
     if (r == null)
        err ("Undefined rule: " + name);
     return r;
  }

  public void storeRule (RuleNode rule, String name)
  {
    if (getRule (name) != null)
       err ("Redefined rule: " + name);

    RuleInfo item = new RuleInfo ();
    item.ident = name;
    item.node = rule;
    rule_list.store (item);
  }

  public Enumeration getRules ()
  {
    return rule_list.elements ();
  }

}
