
import java.util.Enumeration;

class Scanner extends Tool
{
  /* structures, rule names, modes and types */

  private StructInfo defineStruct (String type_name)
  {
     /* find or create type */
     return info.lookupStruct (type_name);
  }

  private void defineSuper (String type_name,
                            String super_name)
  {
     /* find or create type */
     StructInfo t = info.lookupStruct (type_name);

     /* super */
     StructInfo s = info.lookupStruct (super_name);
     if (t.from == null)
     {
        t.from = s;
        s.addSubclass (t);
     }
     else
     {
        if (t.from != s)
           error ("Type " + type_name + " already have super type " + s.ident);
     }
  }

  /* structures */

  public void openStruct(String type_name)
  {
     super.openStruct (type_name);
     defineStruct (type_name);
  }

  public void addSuper (String type_name,
                        String super_name)
  {
     super.addSuper (type_name, super_name);
     defineSuper (type_name, super_name);
  }

  /* rules */

  public void openRule (String name)
  {
     super.openRule (name);

     info.storeRule (rule, name);
     rule.name = name;
  }

  public void ruleParam (int mode,
                         String type_name)
  {
     super.ruleParam (mode, type_name);

     rule.mode = mode;

     StructInfo t = defineStruct (type_name);
     rule.type = t;
     context.type = t;
  }

  public void closeRule ()
  {
     super.closeRule ();
     if (rule.type == null) error ("Internal error");
  }

  /* binary */

  public void openBinary (String branch)
  {
     super.openBinary (branch);
     rule.mode = BINARY_MODE;
     rule.type = defineStruct (info.expr_type);
     defineStruct (info.binary_expr_type);
     defineSuper  (info.binary_expr_type, info.expr_type);
  }

  /* unary */

  public void unary (String key, String tag_value, String branch)
  {
     super.unary (key, tag_value, branch);
     rule.mode = UNARY_MODE;
     defineStruct (info.expr_type);
     rule.type = defineStruct (info.unary_expr_type);
     defineSuper (info.unary_expr_type, info.expr_type);
  }

  /* primitive */

  public void openPrimitive ()
  {
     super.openPrimitive ();
     rule.mode = PRIMITIVE_MODE;
     rule.type = info.string_type;
  }

  /* result options */

  public void resultOption (String type_name)
  {
     super.resultOption (type_name);
     defineStruct (type_name);
     defineSuper (type_name, rule.type.ident);
  }

  /* option */

  private void checkIdent (String value)
  {
     if (! isIdentifier (value))
        error ("Identifier expected: " + value);
  }

  private String checkString (String value)
  {
     if (empty (value) || value.charAt (0) != '"')
        error ("String expected: " + value);

     return value.substring (1, value.length()-1);
  }

  public void option (String name, String value)
  {
     super.option (name, value);

     if (name.equals ("product_name"))
     {
        checkIdent (value);
        info.product_name = value;
     }
     else if (name.equals ("padding"))
     {
        checkIdent (value);
        info.padding = value;
     }
     else if (name.equals ("util_header"))
     {
        info.util_header = checkString (value);
     }
     else if (name.equals ("basic_type"))
     {
        checkIdent (value);
        info.basic_type = value;
     }
     else if (name.equals ("expr_type"))
     {
        checkIdent (value);
        info.expr_type = value;
     }
     else if (name.equals ("expr_tag"))
     {
        checkIdent (value);
        info.expr_tag = value;
     }
     else if (name.equals ("binary_expr_type"))
     {
        checkIdent (value);
        info.binary_expr_type = value;
     }
     else if (name.equals ("binary_expr_left"))
     {
        checkIdent (value);
        info.binary_expr_left = value;
     }
     else if (name.equals ("binary_expr_right"))
     {
        checkIdent (value);
        info.binary_expr_right = value;
     }
     else if (name.equals ("unary_expr_type"))
     {
        checkIdent (value);
        info.unary_expr_type = value;
     }
     else if (name.equals ("unary_expr_param"))
     {
        checkIdent (value);
        info.unary_expr_param = value;
     }
     else error ("Unknown option " + name);
  }
}
