
import java.util.Vector;
import java.util.Enumeration;

class StructInfo extends TypeInfo
{
  public StructInfo from; // super class

  public StructInfo sequence;
  public boolean sequence_element;

  public String tag_name;
  public String tag_value;

  public boolean custom_produce; // true => user defined produce method

  public boolean declared; // true => declaration already written

  public boolean create_func;
  public boolean any_sequence; // sequence in this class or in superclass
  public static final int param_limit = 3; // max. number of parameters

  /* fields */

  private DictTable fields = new DictTable ();

  public FieldInfo getField (String field_name)
  {
    return (FieldInfo) fields.get (field_name);
  }

  public FieldInfo findField (String field_name)
  {
     FieldInfo f = getField (field_name);
     if (f == null)
        err ("Undefined field: " + field_name);
     return f;
  }

  public FieldInfo createField (String field_name)
  {
    if (getField (field_name) != null)
       err ("Field already defined: " + field_name);

    FieldInfo fld = new FieldInfo ();
    fld.ident = field_name;

    fields.store (fld);
    return fld;
  }

  public FieldInfo lookupField (String field_name)
  {
     FieldInfo f = getField (field_name);

     if (f == null && from != null)
        f = from.lookupField (field_name);

     return f;
  }

  public Enumeration getFields ()
  {
    return fields.elements ();
  }

  public int getFieldCount ()
  {
    return fields.size ();
  }

  /* subclasses */

  private Vector subclasses = new Vector ();

  public void addSubclass (StructInfo type)

  {
    subclasses.add (type);
  }

  public Enumeration getSubclasses ()
  {
    return subclasses.elements ();
  }

  public boolean haveSubclasses ()
  {
    return ! subclasses.isEmpty ();
  }

  /* parameters */

  private Vector params = new Vector ();

  public void addParam (FieldInfo f)
  {
    params.add (f);
  }

  public void copyParams (StructInfo s)
  {
    params.addAll (s.params);
  }

  public FieldInfo getParam (int inx)
  {
    return (FieldInfo) params.get (inx);
  }

  public int getParamCount ()
  {
    return params.size ();
  }

}
