
class Tool extends Common
{
  public ParserInfo info;
  public static boolean first_pass = true;

  /* messages */

  private String location ()
  {
     String txt = "";

     if (rule != null)
     {
        txt = "( rule " + rule.name;

        if (current != null)
        {
           txt = txt + ", nonterminal ";
           if (current.kind == LABELED) txt = txt + current.label + ":";
           if (current.kind == ADD) txt = txt + "<add> ";
           if (current.kind == CALL) txt = txt + "<call> ";
           if (current.kind == TMP) txt = txt + "<tmp> ";
           txt = txt + current.subrule;
        }

        txt = txt + ") ";
     }

     return txt;
  }

  public void error (String s)
  {
     s = "Error " + location () + ": " + s;
     err (s);
  }

  public void warning (String s)
  {
     s = "Warning " + location () + ": " + s;

     System.out.println (s);
     System.err.println (s);
  }

  public void trace (String s)
  {
     s = "Trace " + location () + ": " + s;

     System.out.println (s);
  }

  /* rules */

  public RuleNode rule;
  public Context context;

  public final void startRule (RuleNode node, String name)
  {
     rule = node;
     if (first_pass) rule.name = name;

     if (first_pass) rule.context = new Context ();
     context = rule.context;

     openRule (name);
  }

  public final void onRuleParam (int mode,
                                 String type_name)
  {
     if (first_pass) rule.mode = mode;
     ruleParam (mode, type_name);
  }

  public final void stopRule ()
  {
     closeRule ();
     context = null;
     rule = null;
  }

  /* blocks */

  public BlockNode block = null;

  public final void startBlock (BlockNode node)
  {
     node.above = block;
     block = node;
     block.save_alt = alt;

     openBlock ();
     if (is_choose_block ()) openChooseBlock ();
  }

  public final void stopBlock ()
  {
     if (is_choose_block ()) closeChooseBlock ();
     closeBlock ();

     alt = block.save_alt;
     block = block.above;
  }

  public boolean is_select_block ()
  {
   return /* alt.block.above == null && */
          rule.mode == SELECT_MODE;
  }

  public boolean is_main_choose_block ()
  {
   return block.above == null &&
          rule.mode == CHOOSE_MODE;
  }

  public boolean is_choose_block ()
  {
   return block.above != null &&
          block.above.above == null &&
          rule.mode == CHOOSE_MODE;
  }

  /* alternatives */

  public AltNode alt = null;

  public final void startAlternative (AltNode node)
  {
     alt = node;
     alt.block = block;

     if (first_pass)
     {
        alt.save_context = context;
        alt.context = context;
     }
     else
     {
        context = alt.context;
     }

     if (first_pass)
     {
        block.count ++;
        alt.index = block.count;
     }

     // if (is_choose_block ()) openChooseAlternative ();
     openAlternative ();
  }

  public final void stopAlternative ()
  {
     closeAlternative ();
     // if (is_choose_block ()) closeChooseAlternative ();

     context = alt.save_context;
     alt = null;
  }

  /* nonterminal */

  public NontermNode current;

  public final void onNonterminal (NontermNode node,
                                   int kind,
                                   String label,
                                   String name)
  {
     current = node;

     current.kind = kind;
     current.label = label;
     current.subrule = name;

     nonterminal (kind, label, name);

     current = null;
  }

  /* result option */

  public void onResultOption (String type_name)
  {
     if (first_pass)
     {
        context = new Context ();
        alt.context = context;
        rule.context_count ++;
        context.index = rule.context_count;
     }

     resultOption (type_name);
  }

  /* actions */

  public void openStruct (String type_name) { }
  public void closeStruct () { }
  public void structField (String name, String type) { }
  public void structSequence (String name, String type) { }

  public void addTag (String type_name, String tag_name, String tag_value) { }
  public void addSuper (String type_name, String super_name) { }
  public void addCustom (String type_name, String key) { }

  public void openEnum (String name) { }
  public void closeEnum () { }
  public void enumItem (String name) { }

  public void literal (String name, String value) { }

  public void option (String name, String value) { }

  public void openClass (String name) { }
  public void closeClass () { }

  public void openRule (String name) { }
  public void closeRule () { }
  public void ruleParam (int mode, String type_name) { }

  public void openBinary (String branch) { }
  public void closeBinary () { }
  public void binaryItem (String key, String tag_value) { }

  public void unary (String key, String tag_value, String branch) { }

  public void openPrimitive () { }
  public void closePrimitive () { }
  public void primitiveItem (String branch) { }

  public void openChooseBlock () { }
  public void closeChooseBlock () { }
  // public void openChooseAlternative () { }
  // public void closeChooseAlternative () { }

  public void openBlock () { }
  public void closeBlock () { }

  public void openAlternative () { }
  public void closeAlternative () { }

  public void terminal (String name) { }

  public void nonterminal (int kind, String label, String name) { }


  public void openPredicate () { }
  public void closePredicate () { }
  public void predicateNonterminal (String name) { }
  public void predicateTerminal (String name) { }

  public void resultOption (String type_name) { }
  public void storeOption (String name) { }
  public void setOption (String name, String value) { }
  public void silentOption () { }

  public void styleOption (int style) { }
}
