
/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/RIGHTS.html
 */

options
{
  language=Java;
}

class EGLexer extends Lexer;

options
{
	k=2;
	exportVocab=EG1;
  charVocabulary = '\3'..'\377';	
	testLiterals=false;
	// defaultErrorHandler=true; 
     // do not use default error handler in ANTLR 2.7.5 lexer
}

tokens
{
   "struct";
   "enum";
   "literals";
   "options";

   "class";

   "new";
   "modify";
   "return";
   "select";
   "choose";
   "binary";
   "unary";
   "primitive";

   "result";
   "store";
   "add";
   "tmp";
   "call";
   "set";
   "silent";
   "custom";

   "space";
   "nospace";
   "newline";
   "empty";
   "indent";
   "unindent";
}

{
   boolean inside_block;
   /* keywords are disabled inside { ... } */

   boolean inside_class;
   /* keywords are disabled after "class" */

   boolean inside_option;
   /* keywords are temporary enabled at the beginng of < ... > option */

   public int testLiteralsTable (int ttype)
   {
      if ((! inside_class && ! inside_block) || inside_option)
      {
         ttype = super.testLiteralsTable (ttype);

         if (ttype == LITERAL_class)
            inside_class = true;
            // disable keywords after "class" keyword

         if (inside_option)
            inside_option = false;
            // disable keywords after first idenifier in < ... > option
      }
      return ttype;
   }

}

OPEN_OPTION  : '<' { inside_option = true;  };
CLOSE_OPTION : '>' { inside_option = false; };

LCURLY:	       '{' { inside_block = true; };
RCURLY:	       '}' { inside_block = false; };

WS	:	
		(	
			/* '\r' '\n' can be matched in one alternative or by matching
				 '\r' in one iteration and '\n' in another.  I am trying to
				 handle any flavor of newline that comes in, but the language
				 that allows both "\r\n" and "\r" and "\n" to all be valid
				 newline is ambiguous.  Consequently, the resulting grammar
				 must be ambiguous.  I'm shutting this warning off.
			*/
			
			options {	generateAmbigWarnings=false; }	:	
	
				' '
			|	'\t'
			|	'\r' '\n'	{newline();}
			|	'\r'		{newline();}
			|	'\n'		{newline();}
		)
		{ $setType(Token.SKIP); }
		;

COMMENT : 
	( SL_COMMENT | ML_COMMENT) {$setType(Token.SKIP);}
	;

protected
SL_COMMENT :
	"//"
	( ~('\n'|'\r') )*
	(
		/*	'\r' '\n' can be matched in one alternative or by matching
			'\r' and then in the next token.  The language
			that allows both "\r\n" and "\r" and "\n" to all be valid
			newline is ambiguous.  Consequently, the resulting grammar
			must be ambiguous.  I'm shutting this warning off.
		 */
		
		options {	generateAmbigWarnings=false; } :	
		
		  '\r' '\n'
		|	'\r'
		|	'\n'
	)
	{ newline(); }
	;

protected
ML_COMMENT :
	"/*"
	(
		/*	'\r' '\n' can be matched in one alternative or by matching
			'\r' and then in the next token.  The language
			that allows both "\r\n" and "\r" and "\n" to all be valid
			newline is ambiguous.  Consequently, the resulting grammar
			must be ambiguous.  I'm shutting this warning off.
		 */
		
		options
		{
			greedy=false;  // make it exit upon "*/"
			generateAmbigWarnings=false; // shut off newline errors
		}	:	
		
			'\r' '\n'	{ newline(); }
		|	'\r'		  { newline(); }
		|	'\n'		  { newline(); }
		|	~('\n'|'\r')
	)*
	"*/"
	;

LPAREN:	'(' ;

RPAREN:	')' ;

LBRACKET:	'[' ;

RBRACKET:	']' ;

QUESTION :	'?' ;

STAR:	'*' ;

PLUS:	'+' ;

IMPL: "=>" ;

ASSIGN: '=' ;

OR:	'|' ;

SEMI:	';' ;

COMMA: ',' ;

COLON: ':' ;

CHAR_LITERAL:	
	'\'' ( ~ '\'' )* '\''
	;

STRING_LITERAL	:
	'"' ( ~ '"' )* '"'
	;

NUMBER	:	
	('0'..'9')+
	;

NAME
  options { testLiterals = true; }
  :
	('a'..'z' | 'A'..'Z' | '_')
	(	// scarf as many letters/numbers as you can
		options {	warnWhenFollowAmbig=false; } :
		'a'..'z'|'A'..'Z'|'_'|'0'..'9'
	)*
	;
