
options
{
   product_name = CMM; // output file name prefix
   padding = cpp; // input file name prefix
   util_header = "util.h";

   basic_type = TBasic;

   expr_type = TExpr;
   expr_tag = kind;

   binary_expr_type  = TBinaryExpr;
   binary_expr_left  = left;
   binary_expr_right = right;

   unary_expr_type   = TUnaryExpr;
   unary_expr_param  = param;
}

literals
{
   LPAREN     : '(' ;
   RPAREN     : ')' ;
   LBRACKET   : '[' ;
   RBRACKET   : ']' ;
   LBRACE     : '{' ;
   RBRACE     : '}' ;

   DOT        : '.'  ;
   ARROW      : "->" ;
   SCOPE      : "::" ;

   LOG_NOT    : '!' ;
   BIT_NOT    : '~' ;

   INC		    : "++"	;
   DEC		    : "--"	;

   DOT_STAR   : ".*"  ;
   ARROW_STAR : "->*" ;

   STAR       : '*' ;
   SLASH      : '/' ;
   MOD	      : '%' ;

   PLUS       : '+' ;
   MINUS      : '-' ;

   SHL  	    :	"<<"	;
   SHR	      :	">>"	;

   LESS             : '<'  ;
   GREATER          : '>'  ;
   LESS_OR_EQUAL    : "<=" ;
   GREATER_OR_EQUAL : ">=" ;

   EQUAL            : "==" ;
   UNEQUAL          : "!=" ;

   BIT_AND	  : '&' ;
   BIT_XOR	  : '^' ;
   BIT_OR		  : '|' ;

   LOG_AND	  : "&&"	;
   LOG_OR		  : "||"	;

   ASSIGN         : '='   ;
   STAR_ASSIGN		:	"*="	;
   DIV_ASSIGN		  :	"/="	;
   MOD_ASSIGN		  :	"%="	;
   PLUS_ASSIGN		:	"+="	;
   MINUS_ASSIGN	  :	"-="	;
   SHL_ASSIGN	  	:	"<<="	;
   SHR_ASSIGN	  	:	">>="	;
   BIT_AND_ASSIGN	:	"&="	;
   BIT_XOR_ASSIGN	:	"^="	;
   BIT_OR_ASSIGN	:	"|="	;

   QUESTION   : '?' ;
   COMMA      : ',' ;

   COLON      : ':' ;
   SEMICOLON  : ';' ;
   DOTS       : "..." ;
}

/* --------------------------------------------------------------------- */	

enum TAccess
{
  NoAccess,
  PrivateAccess,
  ProtectedAccess,
  PublicAccess
};

struct TMemberSect
{
  TAccess access;
};

struct TBaseItem
{
  TAccess access;
};


/*
enum TStatKind { };

struct TStat
{
  TStatKind tag;
};
*/

/* --------------------------------------------------------------------- */	

class CMMParser;

/* --------------------------------------------------------------------- */	

stat_list <TStatSect>:
   ( <add> stat )* ;

stat <return TStat>:
  <indent>
  internal_stat
  <unindent>
  ;

internal_stat <select TStat>:
  ( declaration ) => declaration_stat |
  labeled_stat |
  case_stat |
  default_stat |
  expression_stat |
  empty_stat |
  compound_stat |
  if_stat |
  switch_stat |
  while_stat |
  do_stat |
  for_stat |
  break_stat |
  continue_stat |
  return_stat |
  goto_stat |
  try_stat |

  into_stat |
  define_stat |
  when_stat |
  call_stat |
  space_stat |
  location_stat ;

declaration_stat <TDeclStat:TStat, tag=DeclSt>:
	 inner_decl:declaration ;

labeled_stat <TLabeledStat:TStat, tag=LabeledSt>:
  lab:identifier ':' body:stat ;

case_stat <TCaseStat:TStat, tag=CaseSt>:
  "case" case_expr:expr ':' body:stat;

default_stat <TDefaultStat:TStat, tag=DefaultSt>:
  "default" ':' body:stat;

expression_stat <TExprStat:TStat, tag=ExprSt>:
   inner_expr:expr ';';

empty_stat <TEmptyStat:TStat, tag=EmptySt>:
	';' ;

compound_stat <TCompoundStat:TStat, tag=CompoundSt>:	
  '{' body:stat_list '}';

if_stat <TIfStat:TStat, tag=IfSt>:
	"if" '(' cond_expr:expr ')' then_stat:stat
	 (
 		<silent>
  	"else" else_stat:stat
	 )? ;

switch_stat <TSwitchStat:TStat, tag=SwitchSt>:
	"switch" '(' switch_expr:expr ')' body:stat;

while_stat  <TWhileStat:TStat, tag=WhileSt>:
	"while" '(' cond_expr:expr ')' body:stat;

do_stat <TDoStat:TStat, tag=DoSt>:
	"do" body:stat "while" '(' cond_expr:expr ')' ';';

for_stat <TForStat:TStat, tag=ForSt>:
	"for"
	'('
  (	
    // [declaration] => from_decl:declaration	|	
    from_expr:expr
	)
	';'
	(to_expr:expr)?	
  ';'
	(cond_expr:expr)?
	')'
	body:stat;

break_stat <TBreakStat:TStat, tag=BreakSt>:
	"break" ';';

continue_stat <TContinueStat:TStat, tag=ContinueSt>:
	"continue" ';';

return_stat <TReturnStat:TStat, tag=ReturnSt>:
	"return" (return_expr:expr)? ';';

goto_stat <TGotoStat:TStat, tag=GotoSt>:
	"goto" goto_lab:identifier ';';

/* --------------------------------------------------------------------- */	

expr <binary>:
	assignment_expr
  ',' CommaExp ;

assignment_expr <binary>:
  conditional_expr // !?
  '='   AssignExp
  "+="  AddAssignExp
  "-="  SubAssignExp
  "*="  MulAssignExp
  "/="  DivAssignExp
  "%="  ModAssignExp
  "<<=" ShlAssignExp
  ">>=" ShrAssignExp
  "&="  AndAssignExp
  "^="  XorAssignExp
  "|="  OrAssignExp ;

conditional_expr <choose TExpr>:
  logical_or_expr
  (
    <result TCondExpr, kind=CondExp>
    <store cond_expr>
    '?' then_expr:expr
    ':' else_expr:conditional_expr
  )? ;

logical_or_expr <binary>:
	logical_and_expr
  "||" LogOrExp ;

logical_and_expr <binary>:
  inclusive_or_expr
  "&&" LogAndExp ;

inclusive_or_expr <binary>:
	exclusive_or_expr
  '|' BitOrExp ;

exclusive_or_expr <binary>:
	and_expr
  '^' BitXorExp ;

and_expr <binary>:
	equality_expr
  '&' BitAndExp ;

equality_expr <binary>:
	relational_expr
  "==" EqExp
  "!=" NeExp ;

relational_expr <binary>:
	shift_expr
  '<' LtExp
  '>' GtExp
  "<=" LeExp
  ">=" GeExp ;

shift_expr <binary>:
	additive_expr
  "<<" ShlExp
  ">>" ShrExp ;

additive_expr <binary>:
	multiplicative_expr
  '+' AddExp
  '-' SubExp ;

multiplicative_expr <binary>:
	unary_expr
  '*' MulExp
  '/' DivExp
  '%' ModExp ;

pm_expr <binary>:
  cast_expr
  ".*"  DotMemberExp
  "->*" ArrowMemverExp ;

cast_expr <return TExpr>:
  unary_expr ; // !?

unary_expr <select TExpr>:
 	  inc_expr
 	|	dec_expr
  | deref_expr
  | addr_expr
  | plus_expr
  | minus_expr
  | bit_not_expr
  | log_not_expr
  | sizeof_expr
  | allocation_expr
  | deallocation_expr
  // | throw_expr // !?
  |	postfix_expr ; /* last - order is important */

inc_expr <unary>:
  "++" PreIncExp unary_expr;

dec_expr <unary>:
  "--" PreDecExp unary_expr;

deref_expr <unary>:
  '*' DerefExp cast_expr;

addr_expr <unary>:
  '&' AddrExp cast_expr;

plus_expr <unary>:
  '+' PlusExp cast_expr;

minus_expr <unary>:
  '-' MinusExp cast_expr;

bit_not_expr <unary>:
  '~' BitNotExp cast_expr;

log_not_expr <unary>:
  '!' LogNotExp cast_expr;

sizeof_expr <unary>:
  "sizeof" SizeofExp unary_expr;

allocation_expr <TNewExpr:TExpr, kind=NewExp>:
  "new"
  (
     type1:qualified_name
  |
    '(' type2:type_description ')'
  )
  (
    '(' init:initializer_list ')'
  )?
  ;

deallocation_expr <TDeleteExpr:TExpr, kind=TDeleteExp>:
  "delete"
  ('[' ']' <set a_array=true>)?
  param:cast_expr ;

postfix_start <select TExpr>:
    modern_cast_expr
  | typeid_expr
  | primary_expr ; /* last - important */

modern_cast_expr <TModernCastExpr:TExpr /* , kind=ModernCastExp */ >:
  ( "dynamic_cast"       <set kind=DynamicCast>
  | "static_cast"        <set kind=StaticCast>
  | "const_cast"         <set kind=ConstCast>
  | "reinterpreter_cast" <set kind=ReinterpreterCast> )
  '<' type:type_description '>'
  '(' param:expr ')'
  ;

typeid_expr <TTypeidExpr:TExpr, kind=TypeidExp>:
  "typeid" '(' param:expr ')';

postfix_expr <choose TExpr>:
  postfix_start
 	(	
    <result TIndexExpr, kind=IndexExp>
    <store left>
    '[' right:expr ']'
  |	
    <result TCallExpr, kind=CallExp>
    <store param>
    '(' list:expr_list ')'
 	|
    <result TFieldExpr, kind=FieldExp>
    <store param>
    '.' ident:identifier
 	|	
    <result TPtrFieldExpr, kind=PtrFieldExp>
    <store param>
    "->" ident:identifier
 	|
    <result TPostIncExpr, kind=PostIncExp>
    <store param>
    "++"
 	|
    <result TPostDecExpr, kind=PostDecExp>
    <store param>
    "--"
 	)* ;

expr_list <TExprList>:
	( <add> assignment_expr (',' <add> assignment_expr)* )? ;

primary_expr <select TExpr>:	
    literal_expr
  | this_expr
  | name_expr
  | subexpr_expr
  ;

literal_expr <TValueExpr:TExpr, kind=ValueExp>:
  value:literal;

literal <primitive>:
    NUMBER
  | CHAR_LITERAL
  | STRING_LITERAL ;

this_expr <TThisExpr:TExpr, kind=ThisExp>:
  "this";

name_expr <return TNameExpr>:
  qualified_name;

subexpr_expr <TSubexprExpr:TExpr, kind=SubexprExp>:
  '(' param:expr ')';

/* --------------------------------------------------------------------- */	

identifier <primitive>:
  IDENT ;

qualified_name <select TNameExpr:TExpr, kind=NameExp>:
   compound_qualified_name
   | simple_qualified_name;
   // | global_qualified_name; // !?

simple_qualified_name <TSimpleNameExpr:TNameExpr, kind=SimpleNameExp>
   <custom produce>:
   ident:identifier;

compound_qualified_name <TCompoundNameExpr:TNameExpr, kind=CompoundNameExp>:
   head:identifier
   '::'
   tail:qualified_name ;

global_qualified_name <TGlobalNameExpr:TNameExpr, kind=GlobalNameExp>:
   '::'
   tail:qualified_name ;

complete_class_name <return TNameExpr>:
  qualified_name ;

/* ---------------------------------------------------------------------- */

declaration_list <TDeclSect>:
  ( <add> declaration <newline> )* ;

declaration <select TDecl>:

    ( decl_specifiers type_description qualified_name '(' ) =>
    function_declaration
  |
    ( decl_specifiers type_description qualified_name ) =>
    variable_declaration
  |
    ( decl_specifiers qualified_name '(' ) =>
    constructor_declaration

  | enum_declaration
  | class_declaration
  | typedef_declaration
  | linkage_specification
  | template_declaration
  | directive
  | empty_declaration ;

/* declaration specifiers */

decl_specifiers <TDeclSpec>:
  (
    "auto"     < set a_auto = true>      |
    "register" < set a_register = true > |
    "static"   < set a_static = true >   |
    "extern"   < set a_extern = true >   |
    "inline"   < set a_inline = true >   |
    "virtual"  < set a_virtual = true >  |
    "explicit" < set a_explicit = true > |
    "const"    < set a_const = true >    |
    "volatile" < set a_volatile = true >
  )* ;

type_description <TTypeDesc>:
  (
    (
      "signed"   <set a_signed   = true> |
      "unsigned" <set a_unsigned = true>
    )?

    (
      "char"     <set a_char     = true> |
      "short"    <set a_short    = true> |
      "int"      <set a_int      = true> |
      "long"     <set a_long     = true> |
      "float"    <set a_float    = true> |
      "double"   <set a_double   = true>
    )

    <set simple=true>
  |
    "void" <set a_void = true >
  |
     name:qualified_name // (param_list:template_arg_list)? // !?
  )

  (<add> ptr_specifier)*
  ;

/* variable */

variable_declaration <TVariableDecl:TDecl, tag=VarDcl>:
  spec:decl_specifiers
  type:type_description
  name:qualified_name
  ('=' init:initializer_item )?
  ';';

/* function */

function_declaration <TFunctionDecl:TDecl, tag=FuncDcl>:
  spec:decl_specifiers
  type:type_description
  name:qualified_name
  '(' param_list:argument_declaration_list ')'
  (
    (init:ctor_initializer)?
    <newline>
    body:compound_stat
    <empty>
    <newline>
  |
    ( '=' pure_init:const_expr )?
    ';' <set empty=true>
  );

/* constructor */

constructor_declaration <TConstructorDecl:TDecl, tag=ConstructorDcl>:
  spec:decl_specifiers
  name:qualified_name
  '(' param_list:argument_declaration_list ')'
  (
    (init:ctor_initializer)?
    <newline>
    body:compound_stat
    <empty>
    <newline>
  |
    ';' <set empty=true>
  );

/* enum */

enum_declaration <TEnumDecl:TDecl>:
  "enum" name:identifier '{' items:enum_list '}' ';' ;

enum_list <TEnumSect>:
  ( <add> enumerator ( ',' <add> enumerator )* )?;

enumerator <TEnumItem>:
  name:identifier ( '=' init:const_expr )? ;

const_expr <return TExpr>:
  conditional_expr ;

/* typedef */

typedef_declaration <TTypeDecl:TDecl>:
  "typedef" spec:decl_specifiers dcl:declarator;

/* extern "..." */

linkage_specification <TExternDecl:TDecl>:
  "extern"
  language:string_literal
  (
    '{' decl_list:declaration_list '}'
  |
    decl:declaration
  );

string_literal <primitive>:
  STRING_LITERAL ;

/* empty declaration */

empty_declaration <TEmptyDecl:TDecl>:
  ';' ;

/* directives */

directive <TDirective:TDecl>:
  "#" <nospace>
  ( "include" include_name:string_literal |
    "ifdef"   ifdef_name:identifier |
    "ifndef"  ifndef_name:identifier |
    "else"    <set else_directive = true> |
    "endif"   <set endif_directive = true> |
    "define"  define_name:identifier );

/* ---------------------------------------------------------------------- */

declarator <TDeclarator>:
  (<add> ptr_specifier)*

  (
    name:dname
  |
    '(' inner_declarator:declarator ')' // <set complex=true>
  )

  (
	  <silent>
    '(' param_list:argument_declaration_list ')'
    // (cv_specifier_list)? // !?
  |
    '[' (lim:expr)? ']' <set array=true>
  )*
  ;

ptr_specifier <TPtrSpecifier>:
  ( '*' <set pointer=true>
  | '&' <set reference=true>
  // | (complete_class_name '::' '*') => complete_class_name '::' '*' // !?
  )
  // (cv_specifier_list)? // !?
  ;

// cv_specifier_list:
//   cv_specifier (cv_specifier)* ;

dname <TDName>:
    '~' ident:identifier
  | name:qualified_name ;

/* parameter declarations */

argument_declaration_list <TParamSect>:
  ( <add> argument_declaration (',' <add> argument_declaration)* )? ;

argument_declaration <TParamItem>:
  type:type_description
  (name:identifier)?
  ('=' init:assignment_expr)? ;

/* initializer */

initializer_item <select TInitItem>:
     simple_initializer_item
   | multi_initializer_item ;

simple_initializer_item <TInitSimpleItem:TInitItem>:
    expr:assignment_expr ;

multi_initializer_item <TInitMultiItem:TInitItem>:
    '{'
    list:initializer_list
    '}' ;

initializer_list <TInitSect>:
    (
      <add> initializer_item
      (',' <add> initializer_item)*
      // (',')? // !?
    )? ;

/* ---------------------------------------------------------------------- */

/* class */

class_declaration <TClassDecl:TDecl, tag=ClassTyp>:
  ( "class"  <set style=ClassStyle>
  | "struct" <set style=StructStyle>
  | "union"  <set style=UnionStyle> )

  name:identifier

  (
    base_list:base_sect
    <newline>
    '{'
    <indent>
    members:member_group
    <unindent>
    '}'
    <empty>
  )?
  ';' ;

/* members */

member_group <TMemberGroup>:
   <add> first_member_list
   ( <add> member_list )* ;

first_member_list <TMemberSect>:
  <set access=NoAccess>
  list:declaration_list ;

member_list <TMemberSect>:
  ( "private"   <set access=PrivateAccess>
  | "protected" <set access=ProtectedAccess>
  | "public"    <set access=PublicAccess> )
  ':'
  list:declaration_list ;

/* base specification */

base_sect <TBaseSect>:
  ( ':' <add> base_specifier ( ',' <add> base_specifier )* )? ;

base_specifier <TBaseItem>:
  ("virtual" <set a_virtual=true> )?

  ( "private"   <set access=PrivateAccess>
  | "protected" <set access=ProtectedAccess>
  | "public"    <set access=PublicAccess>
  |             <set access=NoAccess> )

  name:complete_class_name ;

/* constructor initializer */

ctor_initializer <TConstructorInitializer>:
  ':' <add> mem_initializer (',' <add> mem_initializer)* ;

mem_initializer <TMemberInitializer>:
   name:qualified_name '(' params:expr_list ')' ;

/* ---------------------------------------------------------------------- */

/* template declaration */

template_declaration <TTemplateDecl:TDecl>:
  "template" '<' params:template_param_list '>' body:declaration ;

template_param_list <TTemplateParamSect>:
  <add> template_param ( ',' <add> template_param )* ;

template_param <select TTemplateParam>:
    type_param
  | value_param ;

type_param <TTypeParam:TTemplateParam>:
  "class" name:identifier ;

value_param <TValueParam:TTemplateParam>:
  type:type_description
  name:identifier
  ('=' init:shift_expr)? ; // !?

/* template arguments */

template_arg_list <TTemplateArgSect>:
   '<' <add> template_arg ( ',' <add> template_arg )* '>';

template_arg <TTemplateArg>:
   value:shift_expr; // !?
	
/* ---------------------------------------------------------------------- */

/* namespace */

namespace_definition <TNamespace> :
  "namespace"
  ( name:qualified_name )?
  '{' decl_list:declaration_list '}' ;

namespace_alias_definition <TNamespaceAlias> :
  "namespace" name:identifier '=' value:qualified_name ';' ;

/* using */

using_declaration <TUsingDecl:TDecl, tag=UsingDcl>:
  "using"
  (
    "namespace" ( name1:qualified_name )? name2:identifier
  |
     name:qualified_name
  );

/* ---------------------------------------------------------------------- */

/* try */

try_stat <TTryStat:TStat, tag=TrySt>:
  "try" body:compound_stat handlers:handler_list;

handler_list <THandlerSect>:
  <add> handler (<add> handler)* ;

handler <THandlerItem>:	
  "catch"
  '('
  exception:exception_declaration
  ')'
  body:compound_stat ;

exception_declaration <TExceptionDecl>:
    type:type_description (ident:identifier)?
  | '...' <set dots=true> ;

throw_expr <TThrowExpr:TExpr>: // !?
	"throw" (expr:expr) ? ;

exception_specification <TExceptionSpec>:  // !?
  "throw" '('
  ( <add> type_description (',' <add> type_description)* )?
  ')' ;

/* ---------------------------------------------------------------------- */

group_stat <TGroupStat:TStat, tag=GroupSt>:	
  body:stat_list ;

newline_stat <TNewlineStat:TStat, tag=NewlineSt>:	
  ;

/* ---------------------------------------------------------------------- */

/* into */

into_stat <TIntoStat:TStat, tag=IntoSt> <custom send>:
  "INTO" ( into_param:expr )?
  ( "STORE" store_param:expr )?
  ( def:define_part ) ?

  (
     "CODE" code:stat_list
  |
     "CLASS" cls:class_declaration
  |
     ( "PUBLIC" <set a_public = true> |
       "PRIVATE" <set a_private = true> )?
     "MEMBER" <newline> member:declaration
  )

  <newline>
  "END";

define_stat <TDefineStat:TStat, tag=DefineSt>:
  def:define_part
  "END";

define_part <TDefineSect>:
  "DEFINE"
  ( <add> define_item )* ;

define_item <TDefineItem>:
   name:identifier
   ":"
   ( "IDENT" <set mode=ident_mode> |
     "TYPE"  <set mode=type_mode> |
     "EXPR"  <set mode=expr_mode> |
     "STAT"  <set mode=stat_mode> )
   ( '=' value:expr )?
   ";" ;

when_stat <TWhenStat:TStat, tag=WhenSt> <custom produce>:
	"WHEN" cond_expr:expr "DO"
	then_body:stat_list
	( "OTHERWISE" else_body:stat_list )?
	"END" ;

call_stat <TCallStat:TStat, tag=CallSt> <custom produce>:
  "CALL"
  body:stat_list
  "END" ;

space_stat <TSpaceStat:TStat, tag=SpaceSt> <custom produce>:
  "SPACE" ;

location_stat <TLocationStat:TStat, tag=LocationSt> <custom produce>:
  "LOCATION" '(' place:expr ')' ;

