
/* ANTLR Translator Generator
 * Project led by Terence Parr at http://www.jGuru.com
 * Software rights: http://www.antlr.org/RIGHTS.html
 */

options
{
  language=Java;
}

class EGParser extends Parser;
options
{
	k=2;
	importVocab=EG1;
	exportVocab=EG2;
	defaultErrorHandler=false;
	buildAST=true;
}

tokens
{
  STRUCT_DEF;
  FIELD_DEF;
  SEQUENCE_DEF;

  ENUM_DEF;

  LITERALS_DEF;
  LITERAL_DEF;

  OPTIONS_DEF;
  OPTION_DEF;

	CLASS_DEF;
	RULES_DEF;
	RULE_DEF;
	PARAM_DEF;
	SUPER_DEF;
	TAG_DEF;
	
	NEW_DEF;
	MODIFY_DEF;
	RETURN_DEF;
	SELECT_DEF;
	CHOOSE_DEF;
	
	BINARY_DEF;
	BIN_DEF;
	UNARY_DEF;
	PRIMITIVE_DEF;
	
	BLOCK_DEF;
	ALTERNATIVE_DEF;
	MARK_DEF;
	TERMINAL_DEF;
	NONTERMINAL_DEF;
	LABEL_DEF;
	PREDICATE_DEF;
	
	RESULT_DEF;
	STORE_DEF;
	ADD_DEF;
	CALL_DEF;
	TMP_DEF;
	SET_DEF;
	SILENT_DEF;
	CUSTOM_DEF;
	
	SPACE_DEF;
	NOSPACE_DEF;
	NEWLINE_DEF;
	EMPTY_DEF;
	INDENT_DEF;
	UNINDENT_DEF;
}

grammar
  :
    ( structDef | enumDef | literalsDef | optionsDef )*
		classDef
		EOF !
	;
	
/* ---------------------------------------------------------------------- */
	
id	
	:	
		NAME
	;

string_value	
	:	
		CHAR_LITERAL | STRING_LITERAL
	;

simple_value	
	:	
		// NAME | NUMBER | CHAR_LITERAL | STRING_LITERAL
		NAME
	;

/* ---------------------------------------------------------------------- */
	
structDef
	:
		"struct" !
		param
		LCURLY !
		( fieldDef | sequenceDef )*
		RCURLY !
		( SEMI ! )?
		
		{ ## = #( #[STRUCT_DEF, "STRUCT_DEF"], ##); }
	;

param
  :
    id
    ( superDef )?
    ( tagDef )?
		{ ## = #( #[PARAM_DEF, "PARAM_DEF"], ##); }
  ;

superDef
  :
    COLON! id
		{ ## = #( #[SUPER_DEF, "SUPER_DEF"], ##); }
  ;

tagDef
  :
    COMMA! id ASSIGN! simple_value
		{ ## = #( #[TAG_DEF, "TAG_DEF"], ##); }
  ;

custom
  :
    (
       OPEN_OPTION !
       "custom" !
       id
       CLOSE_OPTION !
       { ## = #( #[CUSTOM_DEF, "CUSTOM_DEF"], ##); }
    )?
  ;


fieldDef
	:
		id id SEMI!
		{ ## = #( #[FIELD_DEF, "FIELD_DEF"], ##); }
	;

sequenceDef
	:
		id L id SEMI!
		{ ## = #( #[SEQUENCE_DEF, "SEQUENCE_DEF"], ##); }
	;

/* ---------------------------------------------------------------------- */
	
enumDef
	:
		"enum" !
		id
		LCURLY !
		( id ( COMMA ! id )* )?
		RCURLY !
		( SEMI ! )?
		
		{ ## = #( #[ENUM_DEF, "ENUM_DEF"], ##); }
	;

/* ---------------------------------------------------------------------- */
	
literalsDef
	:
		"literals" !
		LCURLY !
		( literalDef )*
		RCURLY !
		( SEMI ! )?
		
		{ ## = #( #[LITERALS_DEF, "LITERALS_DEF"], ##); }
	;

literalDef
	:
		id COLON! string_value ( SEMI! )?
		{ ## = #( #[LITERAL_DEF, "LITERAL_DEF"], ##); }
	;

/* ---------------------------------------------------------------------- */
	
optionsDef
	:	
		"options" !
	  LCURLY !
		( optionDef )*
		RCURLY !
		( SEMI ! )?
		
		{ ## = #( #[OPTIONS_DEF, "OPTIONS_DEF"], ##); }
	;

optionDef
  :
		id
		ASSIGN !
		optionValue
		SEMI !
		{ ## = #( #[OPTION_DEF, "OPTION_DEF"], ##); }
	;

optionValue
	:	
	  NAME
	  | STRING_LITERAL
		// | CHAR_LITERAL
		// | NUMBER
	;

/* ---------------------------------------------------------------------- */
	
classDef
	:
		"class" !
		id
		SEMI !
		rules
		
		{ ## = #( #[CLASS_DEF, "CLASS_DEF"], ##); }
		
    //	{ ##.setText ("CLASS_DEF"); }
    //	{ ##.setType (CLASS_DEF); }
	;

rules
  :
    ( rule )*
		{ ## = #( #[RULES_DEF, "RULES_DEF"], ##); }
  ;

rule
	:	
		id
		( normal_cont
		| return_cont
		| select_cont
		| choose_cont
		| binary_cont
		| unary_cont
		| primitive_cont )
		{ ## = #( #[RULE_DEF, "RULE_DEF"], ##); }
	;
	
/* ---------------------------------------------------------------------- */
	
normal_cont
	:	
    OPEN_OPTION !
    ( newDef | modifyDef )?
    param
    CLOSE_OPTION !
    custom

		COLON !
		block
		SEMI !
	;
	
return_cont
	:	
    OPEN_OPTION !
    returnDef
    param
    CLOSE_OPTION !

		COLON !
		return_block
		SEMI !
	;
	
select_cont
	:	
    OPEN_OPTION !
    selectDef
	  param
    CLOSE_OPTION !

		COLON !
		select_block
		SEMI !
	;
	
choose_cont
	:	
    OPEN_OPTION !
    chooseDef
    param
    CLOSE_OPTION !
		
		COLON !
		choose_block
		SEMI !
	;
	
binary_cont
	:	
    OPEN_OPTION !
    "binary" !
    CLOSE_OPTION !
		
		COLON !
		binary_block
		SEMI !
	;
	
unary_cont
	:	
    OPEN_OPTION !
    "unary" !
    CLOSE_OPTION !
		
		COLON !
		unary_block
		SEMI !
	;
	
primitive_cont
	:	
    OPEN_OPTION !
    "primitive" !
    CLOSE_OPTION !
		
		COLON !
		primitive_block
		SEMI !
	;
	
/* ---------------------------------------------------------------------- */
	
newDef
  :
    "new" !
  	{ ## = #[NEW_DEF, "NEW_DEF"]; }
  ;	

modifyDef
  :
    "modify" !
  	{ ## = #[MODIFY_DEF, "MODIFY_DEF"]; }
  ;	

returnDef
  :
    "return" !
  	{ ## = #[RETURN_DEF, "RETURN_DEF"]; }
  ;	

selectDef
  :
    "select" !
  	{ ## = #[SELECT_DEF, "SELECT_DEF"]; }
  ;	

chooseDef
  :
    "choose" !
  	{ ## = #[CHOOSE_DEF, "CHOOSE_DEF"]; }
  ;	

/* ---------------------------------------------------------------------- */
	
return_block
  // one alternative
  :
 		return_alternative
		{ ## = #( #[BLOCK_DEF, "BLOCK_DEF"], ##); }
  ;

return_alternative
  // anonymous nonterminal followed by normal alternative
  :
    ( style_option )*
		anonymous_nonterminal
		( set_option | style_option | ebnf | terminal | nonterminal )*
		{ ## = #( #[ALTERNATIVE_DEF, "ALTERNATIVE_DEF"], ##); }
	;

/* ---------------------------------------------------------------------- */
	
select_block
  :
 		select_alternative ( OR! select_alternative )*
		{ ## = #( #[BLOCK_DEF, "BLOCK_DEF"], ##); }
  ;

select_alternative
  // anonymous nonterminal
  :
    ( predicate )?
		anonymous_nonterminal
		{ ## = #( #[ALTERNATIVE_DEF, "ALTERNATIVE_DEF"], ##); }
	;

/* ---------------------------------------------------------------------- */
	
choose_block
  :
 		( choose_alternative | choose_alternative_temp )
		{ ## = #( #[BLOCK_DEF, "BLOCK_DEF"], ##); }
  ;

choose_alternative
  // anonymous nonterminal followed by parenthesis with optional mark
  :
		anonymous_nonterminal
		choose_ebnf
		{ ## = #( #[ALTERNATIVE_DEF, "ALTERNATIVE_DEF"], ##); }
	;

choose_ebnf
	:	
    // parenthesis with optional mark
	  LPAREN !
		b:choose_block_2
		RPAREN !
		m:ebnf_mark !
		{
		  if (#m != null)
		     ((BlockNode)#b).mark = #m.getText().charAt(0);
		}
	;

choose_alternative_temp
  // <tmp> nonterminal followed by simple parenthesis
  :
		tmp_nonterminal
		choose_ebnf_temp
		{ ## = #( #[ALTERNATIVE_DEF, "ALTERNATIVE_DEF"], ##); }
	;

choose_ebnf_temp
  // simple parenthesis
	:	
	  LPAREN !
		b:choose_block_2
		RPAREN !
	;

choose_block_2
  // parenthesis contents
  :
    ( silent_option )?
 		choose_alternative_2 ( OR! choose_alternative_2 )*
		{ ## = #( #[BLOCK_DEF, "BLOCK_DEF"], ##); }
  ;

choose_alternative_2
  // <result>, <store> followed by normal alternative
  :
    result_option ( store_option )?
		( set_option | style_option | ebnf | terminal | nonterminal )*
		{ ## = #( #[ALTERNATIVE_DEF, "ALTERNATIVE_DEF"], ##); }
	;

/* ---------------------------------------------------------------------- */
	
binary_block
  :
    NAME
    binary_alternative ( binary_alternative )*
		{ ## = #( #[BINARY_DEF, "BINARY_DEF"], ##); }
  ;

binary_alternative
  :
		string_value
		NAME
		{ ## = #( #[BIN_DEF, "BIN_DEF"], ##); }
	;

/* ---------------------------------------------------------------------- */
	
unary_block
  :
		string_value
		NAME
    NAME
		{ ## = #( #[UNARY_DEF, "UNARY_DEF"], ##); }
  ;

/* ---------------------------------------------------------------------- */
	
primitive_block
  :
    NAME
    ( OR ! NAME )*
		{ ## = #( #[PRIMITIVE_DEF, "PRIMITIVE_DEF"], ##); }
  ;

/* ---------------------------------------------------------------------- */
	
block
  :
    ( silent_option )?
 		alternative ( OR! alternative )*
		{ ## = #( #[BLOCK_DEF, "BLOCK_DEF"], ##); }
  ;

alternative
  :
		( set_option | style_option | ebnf | terminal | nonterminal )*
		{ ## = #( #[ALTERNATIVE_DEF, "ALTERNATIVE_DEF"], ##); }
	;

ebnf
	:	
	  LPAREN !
		b:block
		RPAREN !
		m:ebnf_mark !
		{
		  if (#m != null)
		     ((BlockNode)#b).mark = #m.getText().charAt(0);
		}
	;

ebnf_mark
  :
  	(QUESTION | STAR | PLUS )?
	;

terminal
	:
  	string_value	
    { ## = #( #[TERMINAL_DEF, "TERMINAL_DEF"], ##); }
	;

nonterminal
	:	
 	  (element_label | add_option | call_option )
    NAME
    { ## = #( #[NONTERMINAL_DEF, "NONTERMINAL_DEF"], ##); }
	;

	
anonymous_nonterminal
	:	
    NAME
    { ## = #( #[NONTERMINAL_DEF, "NONTERMINAL_DEF"], ##); }
	;

tmp_nonterminal
	:	
 	  tmp_option
    NAME
    { ## = #( #[NONTERMINAL_DEF, "NONTERMINAL_DEF"], ##); }
	;

	
element_label
  :	
		id
		COLON!
		{ ## = #( #[LABEL_DEF, "LABEL_DEF"], ##); }
  ;

predicate
  :
    LPAREN !
    ( anonymous_nonterminal | terminal )*
    RPAREN !
    IMPL !
		{ ## = #( #[PREDICATE_DEF, "PREDICATE_DEF"], ##); }
  ;

/* ---------------------------------------------------------------------- */
	
result_option
  :	
		OPEN_OPTION !
		"result" !
		id
 	  ( tagDef )?
		CLOSE_OPTION !
    { ## = #( #[RESULT_DEF, "RESULT_DEF"], ##); }
	;

store_option
  :	
		OPEN_OPTION !
		"store" !
		id
		CLOSE_OPTION !
    { ## = #( #[STORE_DEF, "STORE_DEF"], ##); }
	;

add_option
  :	
		OPEN_OPTION !
		"add" !
		CLOSE_OPTION !
    { ## = #( #[ADD_DEF, "ADD_DEF"], ##); }
	;

call_option
  :	
		OPEN_OPTION !
		"call" !
		CLOSE_OPTION !
    { ## = #( #[CALL_DEF, "CALL_DEF"], ##); }
	;

tmp_option
  :	
		OPEN_OPTION !
		"tmp" !
		CLOSE_OPTION !
    { ## = #( #[TMP_DEF, "TMP_DEF"], ##); }
	;

set_option
  :	
		OPEN_OPTION !
		"set" !
		id
		ASSIGN !
		simple_value
		CLOSE_OPTION !
    { ## = #( #[SET_DEF, "SET_DEF"], ##); }
	;

silent_option
  :	
		OPEN_OPTION !
		"silent" !
		CLOSE_OPTION !
    { ## = #( #[SILENT_DEF, "SILENT_DEF"], ##); }
	;

/* ---------------------------------------------------------------------- */
	
style_option
  :
    space_option |
    nospace_option |
    newline_option |
    empty_option |
    indent_option |
    unindent_option
  ;

space_option
  :	
		OPEN_OPTION !
		"space" !
		CLOSE_OPTION !
    { ## = #( #[SPACE_DEF, "SPACE_DEF"], ##); }
	;

nospace_option
  :	
		OPEN_OPTION !
		"nospace" !
		CLOSE_OPTION !
    { ## = #( #[NOSPACE_DEF, "NOSPACE_DEF"], ##); }
	;

newline_option
  :	
		OPEN_OPTION !
		"newline" !
		CLOSE_OPTION !
    { ## = #( #[NEWLINE_DEF, "NEWLINE_DEF"], ##); }
	;

empty_option
  :	
		OPEN_OPTION !
		"empty" !
		CLOSE_OPTION !
    { ## = #( #[EMPTY_DEF, "EMPTY_DEF"], ##); }
	;

indent_option
  :	
		OPEN_OPTION !
		"indent" !
		CLOSE_OPTION !
    { ## = #( #[INDENT_DEF, "INDENT_DEF"], ##); }
	;

unindent_option
  :	
		OPEN_OPTION !
		"unindent" !
		CLOSE_OPTION !
    { ## = #( #[UNINDENT_DEF, "UNINDENT_DEF"], ##); }
	;

