
options
{
   product_name = PAS;
   padding = pas;

   expr_type = TExpr;
   expr_tag  = kind;

   binary_expr_type  = TBinaryExpr;
   binary_expr_left  = left;
   binary_expr_right = right;

   unary_expr_type   = TUnaryExpr;
   unary_expr_param  = param;
}

literals
{
   DOT       : '.' ;
   COMMA     : ',' ;
   COLON     : ':' ;
   SEMICOLON : ';' ;
   ARROW     : '^' ;
   RANGE     : '..' ;
   ASSIGN    : ':=' ;

   LPAREN    : '(' ;
   RPAREN    : ')' ;
   LBRACKET  : '[' ;
   RBRACKET  : ']' ;

   STAR      : '*' ;
   SLASH     : '/' ;
   PLUS      : '+' ;
   MINUS     : '-' ;

   EQUAL            : '='  ;
   UNEQUAL          : '<>' ;
   LESS             : '<'  ;
   GREATER          : '>'  ;
   LESS_OR_EQUAL    : '<=' ;
   GREATER_OR_EQUAL : '>=' ;
}

/* ---------------------------------------------------------------------- */

enum TStatTag    { };
enum TTypeTag    { };
enum TStyle      { };
enum TCall       { NoCall };
enum TParamMode  { };
enum TAccess     { };
enum TDeclTag    { };
enum TModuleKind { };

struct TStat
{
   TStatTag tag;
};

struct TType
{
   TTypeTag tag;
};

struct TProcType
{
   TStyle style;
   TCall call;
};

struct TParamItem
{
   TParamMode mode;
};

struct TDecl
{
   bug name;
   TDeclTag tag;
};

struct TProcDecl
{
   TStyle style;
   TCall call;
};

struct TModule
{
  bug name;
  TModuleKind kind;
  TImportSect intf_imports;
  TImportSect impl_imports;
  TDeclGroup intf_decl;
  TDeclGroup impl_decl;
  TStatSeq init;
  TStatSeq finish;
};

/* ---------------------------------------------------------------------- */

class PascalParser;

number <bug> : "number_literal" ; /* !? */
real_number <bug> : "real_literal" ;
identifier <bug> : "ident_literal" ;
string <bug> : "string_literal" ;
label_ident <bug>: "label_identifier" ;

/*
number <literal> : NUM_INT ;
real_number <literal> : NUM_REAL ;
identifier <literal> : IDENT ;
string <literal> : STRING_LITERAL ;

label_ident <select literal>:
    identifier | number;
*/

ifc_type <return TType> :
   simple_alias_type ;

param_type <select TType> :
   simple_alias_type |
   simple_string_type |
   simple_file_type ;

result_type <select TType> :
   simple_alias_type |
   simple_string_type ;

simple_alias_type <new TAliasType> :
   qual_ident:simple_expr ; /* !? */

simple_string_type <new TStringType> :
   "string" ;

simple_file_type <new TFileType> :
   "file" ;

/* ---------------------------------------------------------------------- */

index_expr_item <new TInxItem> :
   ref:expr ;

index_expr_list <new TInxList> :
   <add> index_expr_item ( ',' <add> index_expr_item )* ;

/* arguments */

arg_item <TArgItem> :
   value:expr
   ( ':' width:expr ( ':' digits:expr )? )? ;

arg_list <TArgSect> :
   <add> arg_item ( ',' <add> arg_item )* ;

/* set constructor */

elem_sect <TElemSect> :
   <add> elem_item ( ',' <add> elem_item )* ;

elem_item <TElemItem> :
   low:expr
   (
      '..'
      high:expr
   )? ;

/* structure constructor */

struct_item <TStructSect> :
   ( name:identifier ':' )?
   value:expr ;

struct_sect <TStructSect> :
   <add> struct_item
   (
     ',' <add> struct_item ( ',' <add> struct_item )*
   |
     ';' <add> struct_item ( ';' <add> struct_item )*
   )? ;

/* factor */

ident_expr <TIdentExpr:TExpr> :
   ident:identifier ;

int_value_expr <TIntValueExpr:TExpr> :
   value:number ;

float_value_expr <TFltValueExpr:TExpr> :
   value:real_number ;

string_value_expr <TStrValueExpr:TExpr> :
   value:string ;

nil_expr <TNilExpr:TExpr> :
   'nil' ;

sub_expr <TStructExpr:TExpr> :
   '(' list:struct_sect ')' ;

not_expr <TNotExpr:TExpr> :
   'not' param:expr ;

plus_expr <TPlusExpr:TExpr> :
   '+' param:expr ;

minus_expr <TMinusExpr:TExpr> :
   '-' param:expr ;

adr_expr <TAdrExpr:TExpr> :
   '@' param:expr ;

set_expr <TSetExpr:TExpr> :
   '[' list:elem_sect ']' ;

super_expr <TSuperExpr:TExpr> :
   'inherited' ident:identifier ;

string_type_expr <TStringExpr:TExpr> :
   'string'
   ( "(" param:expr ")" )? ;

variable <select TExpr> :
   ident_expr |
   sub_expr |
   super_expr ;

reference <choose TExpr> :
   variable
   (
      <result TIndexExpr, tag=IndexExp>
      <store left>
      '['
      list:index_expr_list
      ']'
   |
      <result TCallExpr, tag=CallExp>
      <store func>
      '('
      list:arg_list
      ')'
   |
      <result TDerefExpr, tag=DerefExp>
      <store param>
      '^'
   |
      <result TFieldExpr, tag=FieldExp>
      <store param>
      '.'
      ident:identifier
   )*
   ;

factor <select TExpr> :
   int_value_expr |
   float_value_expr |
   string_value_expr |
   not_expr |
   plus_expr |
   minus_expr |
   adr_expr |
   nil_expr |
   set_expr |
   // string_type_expr |  /* !? */
   reference ;

/* term */

term <binary> :
  factor
  '*'   MulExp
  '/'   RDivExp
  'div' DivExp
  'mod' ModExp
  'shl' ShlExp
  'shr' ShrExp
  'and' AndExp
  'as'  AsExp
  ;

/* simple expression */

simple_expr <binary> :
  term
  '+'   AddExp
  '-'   SubExp
  'or'  OrExp
  'xor' XorExp
  ;

/* expression */

expr <binary> : // !? multi
  simple_expr
  '='  EqExp
  '<>' NeExp
  '<'  LtExp
  '>'  GtExp
  '<=' LeExp
  '>=' GeExp
  'in' InExp
  'is' IsExp
  ;

/* ---------------------------------------------------------------------- */

/* goto */

goto_stat <TGotoStat:TStat, tag=GotoSt> :
  "goto" goto_lab:label_ident ;

/* begin end */

begin_stat <TBlockStat:TStat, tag=BeginSt> :
  "begin" body_seq:stat_list "end" ;

stat_list <TStatSeq> :
   <add> stat ( ';' <add> stat )* ;

/* if */

if_stat <TIfStat:TStat, tag=IfSt> :
  "if"
  cond_expr:expr
  "then"
  then_stat:stat
  (
     <silent>
     "else"
     else_stat:stat
  )? ;

/* case */

case_stat <TCaseStat:TStat, tag=CaseSt> :
  "case"
  case_expr:expr
  "of"
  case_list:case_sect
  (
    "else"
    else_seq:stat_list
  )?
  "end" ;

case_sect <TCaseSect> :
  <add> case_item ( ';' <add> case_item )*;

case_item <TCaseItem> :
   sel_list:elem_sect
   ':'
   sel_stat:stat ;

/* while */

while_stat <TWhileStat:TStat, tag=WhileSt> :
  "while" cond_expr:expr "do" body_stat:stat ;

/* repeat */

repeat_stat <TRepeatStat:TStat, tag=RepeatSt> :
  "repeat" body_seq:stat_list "until" until_expr:expr ;

/* for */

for_stat <TForStat:TStat, tag=ForSt> :
  "for"
  var_expr:variable
  ':='
  from_expr:expr
  ( "to" <set incr=true> | "downto" <set incr=false> )
  to_expr:expr
  "do"
  body_stat:stat ;

/* with */

with_stat <TWithStat:TStat, tag=WithSt> :
  "with" with_list:with_sect "do" body_stat:stat ;

with_sect <TWithSect> :
  <add> with_item ( ',' <add> with_item )* ;

with_item <TWithItem> :
  expr:variable ;

/* raise */

raise_stat <TRaiseStat:TStat, tag=RaiseSt> :
  "raise" ( raise_expr:expr )? ;

/* try */

try_stat <TTryStat:TStat, tag=TrySt> :
  "try"
  body_seq:stat_list
  (
    finally_branch:finally_part
  |
    except_branch:except_part
  )
  "end" ;

finally_part <TFinallyPart> :
  "finally"
  finally_seq:stat_list ;

except_part <TExceptPart> :
  "except"
  (
     on_list:on_sect
     ( "else" else_seq:stat_list )?
  |
     else_seq:stat_list
  );

on_sect <TOnSect> :
  <add>  on_item
  ( ';' <add> on_item )*
  ( ';' )? ;

on_item <TOnItem> :
  "on"
  ( on_ident:identifier ':' )?
  on_type:type
  "do"
  body_stat:stat ;

/* empty statement */

empty_stat <TEmptyStat:TStat, tag=EmptySt> : ;

/* simple statement */

simple_stat <choose TStat> :
  <tmp> expr
  (
     <result TAssignStat, tag=AssignSt>
     <store left_expr>
     ':='
     right_expr:expr
  |
     <result TCallStat, tag=CallSt>
     <store call_expr>
  ) ;

/* labeled statement */

labeled_stat <TLabeledTStat:TStat, tag=LabeledSt> :
  lab:label_ident
  ':'
  body:stat ;

/* statement */

stat <select TStat> :
  labeled_stat |
  goto_stat |
  begin_stat |
  if_stat |
  case_stat |
  while_stat |
  repeat_stat |
  for_stat |
  with_stat |
  raise_stat |
  try_stat |
  empty_stat |
  simple_stat ;

/* ---------------------------------------------------------------------- */

param_ident <TParamIdent> :
  name:identifier;

param_item <TParamItem> :
  ( "var"   <set mode=VarParam>   |
    "const" <set mode=ConstParam> |
    "out"   <set mode=OutParam>   |
            <set mode=ValueParam> )

  <add> param_ident ( ',' <add> param_ident )*
  ( ':' typ:param_type ( '=' ini:expr )? )? ;

formal_param_list <TParamSect> :
   '('
       (
          <add> param_item
          ( ';' <add> param_item )*
       )?
   ')';

property_param_list <TParamSect> :
   '['
      <add> param_item
      ( ';' <add> param_item )*
   ']';

/* ---------------------------------------------------------------------- */

class_type <TClassType:TType> :
   "class"
   (
     '('
     parent:ifc_type
     ( ',' ifc_list:interface_sect )?
     ')'
   )?
   items:components_sect
   "end" ;

interface_type <TInterfaceType:TType, tag=InterfaceTyp> :
  "interface"
   ( '(' ifc_list:interface_sect ')' )?
   items:components_sect /* !? */
   "end" ;

interface_sect <TInterfaceSect> :
   <add> interface_item ( ',' <add> interface_item )* ;

interface_item <TIntefaceItem>:
   ifc:ifc_type ;

components_sect <TComponetsSect> :
   <add> components_first_item
   ( <add> components_item )* ;

components_item <TComponentsItem> :
   ( "private"   <set acs=PrivateAccess> |
     "protected" <set acs=ProtectedAccess> |
     "public"    <set acs=PublicAccess> |
     "published" <set acs=PublishedAccess> |
     "automated" <set acs=AutomatedAccess> )
   members:member_sect ;

components_first_item <TComponentsItem> :
   <set acs=PublicAccess>
   members:member_sect ;

member_sect <TDeclGroup> :
   ( <add> member_decl )* ;

member_decl <select TDeclSect> :
   field_decl |
   method_decl |
   property_decl ;

/* field */

field_ident <TVarItem> :
   name:identifier ;

field_decl <TVarDecl:TDeclSect> :
   <add> field_ident
   ( ',' <add> field_ident )*
   ':'
   typ:type;

/* method */

method_decl <return TProcDecl> :
   proc_head
   (
     "virtual" <set a_virtual=true> |
     "dynamic" <set a_dynamic=true> |
     "message" a_message:expr |
     "override" <set a_override=true> |
     "abstract" <set a_abstract=true>
   )* ;

/* property */

property_decl <TPropertyDecl:TDeclSect> :
  "property"
  name:identifier
  ( param_list:property_param_list )?
  ( ':' typ:result_type )?

  ( "index" index:expr )?
  ( "read" a_read:expr )?
  ( "write" a_write:expr )?
  ( "stored" <set a_stored=true> )?
  ( "default" <set a_default=true> )?
  ( "nodefault" <set a_nodefault=true> )?
  ( "implements" a_implements:interface_sect )?

  ( ';' "default" ';' <set a_def_array=true> )?
  ';' ;

/* ---------------------------------------------------------------------- */

/* enum */

enum_type <TEnumType:TType, tag=EnumTyp> :
   '(' elements:enum_sect ')' ;

enum_sect <TEnumSect> :
   <add> enum_item ( ',' <add> enum_item )* ;

enum_item <TEnumItem> :
   name:identifier ;

/* string */

string_type <TStringType:TType, tag=StringTyp> :
   "string"
   ( '[' lim:expr ']' )? ;

/* array */

array_type <TArrayType:TType, tag=ArrayTyp> :
  "array"
  ( '[' index_list:index_type_sect ']' )?
  "of"
  elem:type;

index_type_sect <TIndexTypeSect> :
   <add> index_type_item ( ',' <add> index_type_item )* ;

index_type_item <TIndexTypeItem> :
   index:type ;

/* record */

record_type <TRecordType:TType, tag=RecordTyp> :
  "record"
  fields:field_sect
  "end"
  ;

field_sect <TDeclGroup> :
   ( <add> field_decl )* ;

/* pointer */

pointer_type <TPointerType:TType, tag=PointerTyp> :
   '^' elem:type;

/* set */

set_type <TSetType:TType, tag=SetTyp> :
   "set" "of" elem:type;

/* file */

file_type <TFileType:TType, tag=FileTyp> :
   "file" ( "of" elem:type )? ;

/* class of type */

class_of_type <TClassOfType:TType, tag=ClassOfTyp> :
  "class" "of" elem:type ;

/* procedure */

proc_type <TProcType:TType, tag=ProcTyp> :
   (
     "procedure" <set style = ProcedureStyle> |
     "function"  <set style = FunctionStyle>
   )

   param_list:formal_param_list
   ( ':' answer:result_type )?
   ( "of" "object" <set of_object=true> )?

   (
     "register" <set call=RegisterCall> |
     "pascal"   <set call=PascalCall> |
     "cdecl"    <set call=CdeclCall> |
     "stdcall"  <set call=StdcallCall> |
     "safecall" <set call=SafecallCall>
   )? ;

/* other type */

other_type <choose TType> :
   <tmp> simple_expr
   (
     <result TSubrangeType, tag=SubrangeTyp>
     <store low>
     '..'
     high:simple_expr
   |
     <result TAliasType, tag=AliasTyp>
     <store qual_ident>
   );

/*
range_type <TSubrangeType:TType, tag=SubrangeTyp> :
   low:simple_expr
   '..'
   high:simple_expr ;

alias_type <TAliasType:TType, tag=AliasTyp> :
   qual_ident:simple_expr ;
*/

/* type */

type <select TType> :
   string_type |
   array_type |
   set_type |
   file_type |
   record_type |
   class_type |
   class_of_type |
   interface_type |
   proc_type |
   pointer_type |
   // enum_type | /* !? */
   other_type ;

   /* !? packed */

/* ---------------------------------------------------------------------- */

/* labels */

label_decl <TLabelDecl, tag=LabelDcl> :
   name:label_ident ;

label_sect <TLabelSect:TDeclSect> :
   "label" <add> label_decl ( ',' <add> label_decl )* ';' ;

/* constants */

const_decl <TConstDecl, tag=ConstDcl> :
   name:identifier ( ':' typ:type )? '=' val:expr ';' ;

const_sect <TConstSect:TDeclSect> :
   "const" <add> const_decl ( <add> const_decl )* ;

/* types */

type_decl <TTypeDecl, tag=TypeDcl> :
   name:identifier '=' typ:type ';' ;

type_sect <TTypeSect:TDeclSect> :
   "type" <add> type_decl ( <add> type_decl )* ;

/* variables */

var_item <TVarItem> :
   name:identifier ;

var_decl <TVarDecl, tag=VarDcl> :
   <add> var_item ( ',' <add> var_item )*
   ':' typ:type
   ( '=' ini:expr )?
   ';' ;

var_sect <TVarSect:TDeclSect> :
   "var" <add> var_decl ( <add> var_decl )* ;

/* subroutine */

proc_head <TProcDecl:TDeclSect, tag=ProcDcl> :

   ( "class" <set a_static=true> )?

   ( "procedure"   <set style=ProcedureStyle>   |
     "function"    <set style=FunctionStyle>    |
     "constructor" <set style=ConstructorStyle> |
     "destructor"  <set style=DestructorStyle>  )

   ( context:identifier ':' )?
   name:identifier
   param_list:formal_param_list
   ( ':' answer:result_type ) ?

   (
     "reintroduce" <set a_reintroduce=true> |
     "overload" <set a_overload=true>
   )?

   (
     ( ';' )?
     ( "register" <set call=RegisterCall> |
       "pascal"   <set call=PascalCall>   |
       "cdecl"    <set call=CdeclCall>    |
       "stdcall"  <set call=StdcallCall>  |
       "safecall" <set call=SafecallCall> )
   )?

   ';' ;

proc_decl <return TProcDecl> :
   proc_head
   ( "forward" <set a_forward=true>
   | <call> proc_external
   | local:decl_part
     "begin"
     body:stat_list
     "end"
   )
   ';' ;

proc_external <modify TProcDecl> :
   "external" <set a_external=true>
   (
      a_lib:expr
      (
         "name"
         a_name:expr
      |
         "index"
         a_index:expr
      )?
   )?
   ';' ;

proc_intf_decl <return TProcDecl> :
   proc_head
   ( <call> proc_external )? ;

/* declarations */

intf_decl <select TDeclSect> :
   const_sect |
   type_sect |
   var_sect |
   proc_intf_decl ;

intf_decl_part <TDeclGroup> :
   ( <add> intf_decl )* ;

decl <select TDeclSect> :
   label_sect |
   const_sect |
   type_sect |
   var_sect |
   proc_decl ;

decl_part <TDeclGroup> :
   ( <add> decl )* ;

/* import */

import_item <new TImportItem> :
  name:identifier
  ( "in" path:string )? ;

import_sect <new TImportSect> :
  (
     "uses"
     <add> import_item
     ( ',' <add> import_item )*
     ';'
  )? ;

/* unit */

unit_decl <new TUnitModule:TModule, kind=UnitMod> :
   "unit" name:identifier ';'
   "interface"
      intf_imports:import_sect
      intf_decl:intf_decl_part
   "implementation"
      impl_imports:import_sect
      impl_decl:decl_part
   (
     ( "initialization" init:stat_list )?
     ( "finatlization" finish:stat_list )?
   |
     "begin" init:stat_list
   )
   "end" '.'
   ;

/* program */

program_decl <new TProgramModule:TModule, kind=ProgramMod> :
   "program" name:identifier ';'
   intf_imports:import_sect
   intf_decl:decl_part
   "begin"
   init:stat_list
   "end" '.' ;

/* library */

library_decl <new TLibraryModule:TModule, kind=LibraryMod> :
   "library" name:identifier ';'
   intf_imports:import_sect
   intf_decl:decl_part
   "begin"
   init:stat_list
   "end" '.' ;

/* module */

module_decl <select TModule> :
   unit_decl |
   program_decl |
   library_decl ;
