
options
{
	language="Java";
}

class EGTreeParser extends TreeParser;

options
{
	importVocab=EG2;
	exportVocab=EG3;
	defaultErrorHandler=false;
}

{
   public Tool tool;
   private String tmp_type;
   private int tmp_mode;
   private String tmp_label;
   private int tmp_kind;
}

grammar	
	:
    ( structDef | enumDef | literalsDef | optionsDef )*
    classDef
	;

/* ---------------------------------------------------------------------- */

structDef
	:
	#( STRUCT_DEF
	
	   param
     { tool.openStruct (tmp_type); } /* tmp_type from param */
     // custom /* tmp_type from param */

	   ( fieldDef | sequenceDef )*
	
     { tool.closeStruct (); }
	 )
	;

param
  :
 	#( PARAM_DEF
 	
 	   name:NAME
 	   { tmp_type = #name.getText (); }
 	
 	   ( superDef )?
 	   ( tagDef )?
   )
  ;

superDef
 	:
 	#( SUPER_DEF
 	   name:NAME
 	   { tool.addSuper (tmp_type, #name.getText ()); }
 	 )
 	;

tagDef
 	:
 	#( TAG_DEF
 	   name:NAME
 	   value:simple_value
 	   { tool.addTag (tmp_type, #name.getText (), #value.getText ()); }
 	 )
 	;

custom
  :
  (
     #( CUSTOM_DEF
        name:NAME
        { tool.addCustom (tmp_type, #name.getText ()); }
      )
  )?
  ;

fieldDef
	:
	#( FIELD_DEF
	   type:NAME
	   name:NAME
     { tool.structField (#name.getText (), #type.getText()); }
	 )
	;

sequenceDef
	:
	#( SEQUENCE_DEF 	
	   type:NAME
	   name:NAME
     { tool.structSequence (#name.getText (), #type.getText()); }
   )
	;

/* ---------------------------------------------------------------------- */
	
enumDef
	:
	#( ENUM_DEF
	   name:NAME
     { tool.openEnum (#name.getText()); }

     (
	      fld:NAME
        { tool.enumItem (#fld.getText()); }
	   )*
	
     { tool.closeEnum (); }
	 )
	;

/* ---------------------------------------------------------------------- */
	
literalsDef
  :
  #(LITERALS_DEF (literalDef)* )
  ;

literalDef
  :
  #( LITERAL_DEF
     name:NAME
     value:string_value
     { tool.literal (#name.getText (), Common.strip (#value.getText ())); }
   )
  ;

/* ---------------------------------------------------------------------- */
	
optionsDef
  :
  #(OPTIONS_DEF (optionDef)* )
  ;

optionDef
  :
  #( OPTION_DEF
     name:NAME
     value:simple_value
     { tool.option (#name.getText (), #value.getText ()); }
   )
  ;

string_value
  :
  STRING_LITERAL | CHAR_LITERAL
  ;

simple_value
  :
  NAME | INT | STRING_LITERAL | CHAR_LITERAL
  ;

/* ---------------------------------------------------------------------- */
	
classDef
	:
	#( CLASS_DEF
	   name:NAME
     { tool.openClass (#name.getText()); }
	   rules
     { tool.closeClass (); }
	 )
	;
	
rules
  :
  #(RULES_DEF (rule)* )
  ;

rule
  :
  #( RULE_DEF

     name:NAME

     { tool.startRule ((RuleNode) #rule, #name.getText()); }

     ( binary_block |
       unary_block |
       primitive_block |

       { tmp_mode = Common.NEW_MODE; }
   	   ( NEW_DEF |
   	     MODIFY_DEF { tmp_mode = Common.MODIFY_MODE; } |
   	     RETURN_DEF { tmp_mode = Common.RETURN_MODE; } |
   	     SELECT_DEF { tmp_mode = Common.SELECT_MODE; } |
   	     CHOOSE_DEF { tmp_mode = Common.CHOOSE_MODE; } )?

        param
        { tool.onRuleParam (tmp_mode, tmp_type); } /* tmp_type from param */
        custom /* tmp_type from param */

        block
     )

     { tool.stopRule (); }
   )
  ;

/* ---------------------------------------------------------------------- */
	
binary_block
  :
  #( BINARY_DEF
     branch:NAME
     { tool.openBinary (#branch.getText()); }
     (binary_alternative)*
     { tool.closeBinary (); }
   )  	
  ;

binary_alternative
  :
  #( BIN_DEF
     key:string_value
     tag_value:NAME
     { tool.binaryItem (Common.strip (#key.getText()), #tag_value.getText()); }
   )
  ;

/* ---------------------------------------------------------------------- */
	
unary_block
  :
  #( UNARY_DEF
     key:string_value
     tag_value:NAME
     branch:NAME
     { tool.unary (Common.strip (#key.getText()),
                                 #tag_value.getText(),
                                 #branch.getText()); }
   )  	
  ;

/* ---------------------------------------------------------------------- */
	
primitive_block
  :
  #( PRIMITIVE_DEF
     { tool.openPrimitive (); }
     (primitive_alternative)*
     { tool.closePrimitive (); }
   )  	
  ;

primitive_alternative
  :
    name:NAME
    { tool.primitiveItem (#name.getText()); }
  ;

/* ---------------------------------------------------------------------- */
	
block
  :
  #( BLOCK_DEF
     { tool.startBlock ((BlockNode) #block); }
     (silent_option)?
     (alternative)*
     { tool.stopBlock (); }
   )  	
  ;

alternative
  :
  #( ALTERNATIVE_DEF
     { tool.startAlternative ((AltNode) #alternative); }
     (result_option)?
     (store_option)?
     (predicate)?
     (set_option | style_option | ebnf | terminal | nonterminal)*
     { tool.stopAlternative (); }
   )  	
  ;

ebnf
  :
  block
  ;

terminal
  :
  #( TERMINAL_DEF
     val:string_value
     { tool.terminal (Common.strip (#val.getText ())); }
   )
  ;

nonterminal
  :
  #( NONTERMINAL_DEF
     { tmp_label = ""; }

     ( element_label { tmp_kind = Common.LABELED; } |
       ADD_DEF       { tmp_kind = Common.ADD; }     |
       CALL_DEF      { tmp_kind = Common.CALL; }    |
       TMP_DEF       { tmp_kind = Common.TMP; }     |
                     { tmp_kind = Common.ANONYMOUS; } )

     name:NAME

     { tool.onNonterminal ((NontermNode) #nonterminal,
                           tmp_kind, tmp_label, #name.getText ()); }
   )
  ;

element_label
  :
  #( LABEL_DEF
     name:NAME
     { tmp_label = #name.getText (); }
   )
  ;

/* ---------------------------------------------------------------------- */
	
predicate
  :
  #( PREDICATE_DEF
     { tool.openPredicate (); }
     ( predicate_nonterminal | predicate_terminal )*
     { tool.closePredicate (); }
   )
  ;

predicate_nonterminal
  :
  #( NONTERMINAL_DEF
     name:NAME
     { tool.predicateNonterminal (#name.getText ()); }
   )
  ;

predicate_terminal
  :
  #( TERMINAL_DEF
     val:string_value
     { tool.predicateTerminal (Common.strip (#val.getText ())); }
   )
  ;

/* ---------------------------------------------------------------------- */
	
result_option
  :
  #( RESULT_DEF
     name:NAME

     { tmp_type = #name.getText (); }
     { tool.onResultOption (tmp_type); }

     /* tagDef need tmp_type */
	   ( tagDef )?
   )
  ;

store_option
  :
  #( STORE_DEF
     name:NAME
     { tool.storeOption (#name.getText ()); }
   )
  ;

set_option
  :
  #( SET_DEF
     name:NAME
     value:simple_value
     { tool.setOption (#name.getText (), #value.getText ()); }
   )
  ;

silent_option
  :
  #( SILENT_DEF
     { tool.silentOption (); }
   )
  ;

/* ---------------------------------------------------------------------- */
	
style_option
  :
    space_option |
    nospace_option |
    newline_option |
    empty_option |
    indent_option |
    unindent_option
  ;

space_option
  :
  #( SPACE_DEF
     { tool.styleOption (Common.SPACE); }
   )
  ;

nospace_option
  :
  #( NOSPACE_DEF
     { tool.styleOption (Common.NOSPACE); }
   )
  ;

newline_option
  :
  #( NEWLINE_DEF
     { tool.styleOption (Common.NEWLINE); }
   )
  ;

empty_option
  :
  #( EMPTY_DEF
     { tool.styleOption (Common.EMPTY); }
   )
  ;

indent_option
  :
  #( INDENT_DEF
     { tool.styleOption (Common.INDENT); }
   )
  ;

unindent_option
  :
  #( UNINDENT_DEF
     { tool.styleOption (Common.UNINDENT); }
   )
  ;
